from __future__ import annotations

import asyncio
import atexit
import logging
import os
import socket
import sys
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from enum import StrEnum, auto
from functools import cache, partial, wraps
from pathlib import Path
from typing import (
    Any,
    Concatenate,
    Literal,
    ParamSpec,
    Self,
    TypeVar,
    assert_never,
    cast,
)

import accelerate
import torch
import torch.distributed as dist
import tree
from dask import config
from dask.distributed import Actor, ActorFuture, Client
from distributed.utils import sync
from pydantic import BaseModel, ConfigDict, Field, field_validator
from torch import nn
from torch.cuda import nccl
from torch.distributed.fsdp import (
    BackwardPrefetch,
    FullStateDictConfig,
    FullyShardedDataParallel,
    MixedPrecision,
    ShardingStrategy,
    StateDictType,
)
from torch.nn.functional import pad
from transformers import GenerationMixin, PreTrainedModel, PreTrainedTokenizer
from transformers.generation.utils import GenerateDecoderOnlyOutput, LogitsProcessorList
from transformers.tokenization_utils_base import BatchEncoding

from ldp.graph.async_torch import AsyncBufferedWorker, AsyncTorchModule
from ldp.nn.generation import LogitsProcessorWithFinalize
from ldp.nn.handlers.chunking import TensorChunker, TOutputType
from ldp.nn.handlers.module_handler import ModuleExecutionInterface, ModuleHandler
from ldp.nn.lm_config import LMConfig, TorchDType
from ldp.nn.utils import REPO_ROOT, set_seed

if sys.version_info >= (3, 12):
    from typing import overload
else:
    from typing_extensions import overload  # noqa: UP035

logger = logging.getLogger(__name__)

config.set({
    # We have no use for rebooting workers in aviary for now, and rebooting workers
    # is annoying when debugging.
    "distributed.scheduler.allowed-failures": 0,
    # FSDP forward/backward passes can take way longer than the default warning at 3s
    "distributed.admin.tick.limit": "30s",
    # Gives us more time to debug a downed worker. TODO: see if there are negative consequences
    # of having this always enabled
    "distributed.comm.timeouts.connect": "300s",
    "distributed.comm.timeouts.tcp": "1200s",
})

TReturn = TypeVar("TReturn")
TParams = ParamSpec("TParams")


def is_conversation(messages) -> bool:
    """Check if messages is an instance of Conversation."""
    return isinstance(messages, list) and all(
        isinstance(msg, dict)
        and all(
            isinstance(key, str) and isinstance(value, str)
            for key, value in msg.items()
        )
        for msg in messages
    )


def get_unused_port() -> int:
    """Find an unused port by creating a temporary socket."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        return s.getsockname()[1]


class ExecutionMode(StrEnum):
    LOCAL_MACHINE = auto()
    SLURM_CLUSTER = auto()


class FSDPConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    offload_cpu: bool = Field(
        default=False,
        description="Whether to offload model parameters and gradients to CPU.",
    )
    activation_checkpointing: bool = False
    cpu_ram_efficient_loading: bool = False
    state_dict_type: StateDictType = StateDictType.FULL_STATE_DICT
    backward_prefetch: BackwardPrefetch | None = Field(
        BackwardPrefetch.BACKWARD_PRE,
        description=(
            "See https://pytorch.org/docs/stable/fsdp.html#torch.distributed.fsdp.BackwardPrefetchDefault"
            " is PRE to be consistent with FSDP's default."
        ),
    )

    @field_validator("backward_prefetch", mode="before")
    @classmethod
    def validate_backward_prefetch(cls, v) -> BackwardPrefetch | None:
        if v is None:
            return None
        if isinstance(v, str):
            # BackwardPrefetch isn't a StrEnum, so Pydantic won't cast a string
            return BackwardPrefetch[v]
        return BackwardPrefetch(v)


class ParallelModeConfig(FSDPConfig):
    num_workers: int = Field(description="Number of workers to be used in the cluster.")
    num_cpus_per_worker: int = Field(
        default=1, description="Number of CPUs to be allocated per worker."
    )
    execution_mode: ExecutionMode = Field(
        description=(
            "Execution mode of the current setup, defines how to allocate resources "
            "(cpus/gpus) for the model to run on."
        ),
        default=ExecutionMode.LOCAL_MACHINE,
    )

    scheduler_addr: str = "localhost"
    scheduler_port: int = Field(default=0, description="0 means Dask picks randomly.")
    torch_port: int = Field(default_factory=get_unused_port)

    # Below configurations (walltime, memory, log_directory) are only for ExecutionMode.SLURM_CLUSTER:
    walltime: str = Field(
        default="00:30:00", description="Max time the worker can run."
    )
    memory: str = Field(default="32GB", description="Memory allocated per worker.")
    log_directory: str = Field(
        default=f"{REPO_ROOT}/logs/slurm_outputs/",
        description="Directory to store logs.",
    )


class LMType(StrEnum):
    GENERATION = auto()
    REGRESSION = auto()


class TransformerHandlerConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    lm_config: LMConfig
    lm_type: LMType
    checkpoint: str | None = None

    batch_size: int
    max_wait_interval: float = 0.1
    module_call_fn: Callable
    collate_fn: Callable
    decollate_fn: Callable

    parallel_mode_config: ParallelModeConfig | None = Field(
        default=None,
        description=(
            "Optional configuration for distributing the transformer across "
            "multiple devices/nodes. If not provided, will default to single-device."
        ),
    )

    def make_async_module(self, **kwargs) -> AsyncTransformerInterface:
        if self.parallel_mode_config:
            return ParallelAsyncTransformer(config=self, **kwargs)
        return AsyncTransformer(config=self, **kwargs)


class AsyncTransformerInterface(ModuleExecutionInterface, AsyncTorchModule, ABC):
    """Base class for async interactions with a transformer model."""

    @abstractmethod
    async def __call__(  # type: ignore[override]
        self,
        inputs: str | BatchEncoding | list[dict],
        tools_json: list[dict[Any, Any]] | None = None,
        **kwargs,
    ) -> tuple[str, torch.Tensor]:
        """Call the transformer on a single input, which may be encoded."""

    @staticmethod
    def model_generate(model: PreTrainedModel, *args, **kwargs):
        """A method that can be used as module_call_fn to sample from an LLM."""
        # Summoning params per https://github.com/pytorch/pytorch/issues/100069
        # If model is not FSDP, this context manager is a no-op.
        with FullyShardedDataParallel.summon_full_params(model, recurse=False):
            logger.debug(
                f"model.generate() input_ids shape: {kwargs['input_ids'].shape}, rank"
                f" {os.environ.get('RANK')}"
            )
            if not isinstance(model, GenerationMixin):  # type: ignore[unreachable]
                raise TypeError(
                    "model_generate only supports models that inherit from GenerationMixin"
                )
            return model.generate(  # type: ignore[unreachable]
                *args,
                **kwargs,
                pad_token_id=model.config.pad_token_id,  # not always set properly by .generate()
                eos_token_id=model.config.eos_token_id,
            )


class TransformerHandler(ModuleHandler):
    def __init__(self, config: TransformerHandlerConfig):
        # Maybe this should be configurable? Hard to isolate the effect though
        torch.set_float32_matmul_precision("high")

        self.config = config
        # Use local_rank to resolve model location only in the main process *for each node*
        config.lm_config.resolve_model_location(is_main_process=self.local_rank == 0)

        match config.lm_type:
            case LMType.GENERATION:
                tokenizer, model = config.lm_config.get_causal_lm()
                # On left for https://github.com/huggingface/transformers/pull/7552
                # ^ that seems to work for most HF models w/ absolute position embeddings
                # Left padding always works for relative position embeddings
                tokenizer.padding_side = "left"
            case LMType.REGRESSION:
                tokenizer, model = config.lm_config.get_regression_lm()
            case _:
                assert_never(config.lm_type)
        super().__init__(model)
        self.tokenizer = tokenizer
        maybe_set_tokenizer_chat_template(
            self.tokenizer, self.config.lm_config.chat_template
        )

        self._setup_accelerator()

        if config.checkpoint is not None:
            self.load_checkpoint(config.checkpoint)

    def _setup_accelerator(self):
        self.accelerator = accelerate.Accelerator(
            # This has to be disabled because accelerator wraps forward() to upcast outputs to fp32. That
            # causes problems with generation, where the cache is expected to be in the same dtype as the model.
            # TODO: understand why this doesn't break in huggingface code and re-enable in ours.
            # mixed_precision=("bf16" if self.config.lm_config.dtype == TorchDType.bf16 else "no")
        )
        self.module = self.accelerator.prepare(self.module)

    @property
    def local_rank(self) -> int:
        return int(os.getenv("LOCAL_RANK", "0"))

    def load_checkpoint(self, ckpt: os.PathLike | str, **kwargs) -> None:
        logger.info(f'Loading checkpoint from "{ckpt}"')
        self.accelerator.load_state(str(ckpt), **kwargs)

    def save_checkpoint(self, ckpt: os.PathLike | str, **kwargs) -> None:
        self.accelerator.save_state(str(ckpt), **kwargs)
        self.barrier()
        # We do not want to save random states - they would be loaded by load_state
        # automatically. Clean up after all processes have saved.
        if int(os.getenv("RANK", "0")) == 0:
            random_states = Path(ckpt).rglob("random_state*.pkl")
            for random_state in random_states:
                random_state.unlink()

    @staticmethod
    def barrier() -> None:
        if dist.is_initialized():
            dist.barrier()


class AsyncTransformer(TransformerHandler, AsyncTransformerInterface):
    """Equivalent of AsyncTorchModule, but for a transformer model.

    The model is instantiated in the current process.
    """

    def __init__(self, config: TransformerHandlerConfig):
        # First init the local module
        TransformerHandler.__init__(self, config)
        # Then init the logic to buffer inputs to the module
        AsyncTorchModule.__init__(
            self,
            module=self.module,
            batch_size=config.batch_size,
            max_wait_interval=config.max_wait_interval,
            collate_fn=config.collate_fn,
            decollate_fn=config.decollate_fn,
            module_call_fn=config.module_call_fn,
        )

    async def __call__(
        self,
        inputs: str | BatchEncoding | dict | list[dict] | None = None,
        tools_json: list[dict[Any, Any]] | None = None,
        input_ids: torch.Tensor | None = None,
        attention_mask: torch.Tensor | None = None,
        **kwargs,
    ) -> tuple[str, torch.Tensor]:
        if inputs is None:
            inputs = {"input_ids": input_ids, "attention_mask": attention_mask}
        inputs_tokenized = _get_tokenized_inputs(self.tokenizer, inputs, tools_json).to(
            device=_get_data_device()
        )
        inputs_len = inputs_tokenized["input_ids"].shape[1]
        self.module.eval()
        outputs = await AsyncTorchModule.__call__(self, **inputs_tokenized, **kwargs)
        self._maybe_finalize_logits_processors(kwargs.get("logits_processor"), outputs)

        return _process_outputs(self.config, self.tokenizer, outputs, inputs_len)

    @staticmethod
    def _maybe_finalize_logits_processors(
        logits_processors: LogitsProcessorList | None,
        outputs: GenerateDecoderOnlyOutput,
    ) -> None:
        if not logits_processors:
            return
        for processor in logits_processors:
            if isinstance(processor, LogitsProcessorWithFinalize):
                processor.finalize(outputs.sequences)


class ParallelWorkerConfig(FSDPConfig):
    model_config = ConfigDict(extra="ignore")

    rank: int = Field(description="Rank of the current process.")
    world_size: int = Field(description="Total number of processes.")
    local_rank: int = Field(description="Local rank of the current process.")
    master_addr: str = Field(description="Address of the master node.")
    master_port: int = Field(description="Port of the master node.")

    def set_env_vars(self):
        # These inform torch.distributed how to set up the process group
        os.environ["CUDA_VISIBLE_DEVICES"] = str(self.local_rank)
        os.environ["RANK"] = str(self.rank)
        os.environ["WORLD_SIZE"] = str(self.world_size)
        os.environ["LOCAL_RANK"] = str(self.local_rank)
        os.environ["MASTER_ADDR"] = self.master_addr
        os.environ["MASTER_PORT"] = str(self.master_port)

        # These tell PretrainedModel.from_pretrained() that we're using FSDP
        os.environ["ACCELERATE_USE_FSDP"] = "1"
        os.environ["FSDP_CPU_RAM_EFFICIENT_LOADING"] = str(
            int(self.cpu_ram_efficient_loading)
        )


class ParallelTransformerHandler(TransformerHandler):
    def __init__(
        self,
        config: TransformerHandlerConfig,
        parallel_worker_config: ParallelWorkerConfig,
    ):
        parallel_worker_config.set_env_vars()
        dist.init_process_group(backend="nccl")
        self.worker_config = parallel_worker_config
        super().__init__(config)

    def _setup_accelerator(self):
        bf16 = self.config.lm_config.dtype == TorchDType.bf16

        mixed_precision = None
        if bf16:
            bf16_ready = (
                torch.version.cuda
                and torch.cuda.is_bf16_supported()
                and torch.version.cuda >= "11.0"
                and dist.is_nccl_available()
                and nccl.version() >= (2, 10)
            )
            assert bf16_ready, (
                "Mixed precision training requires CUDA 11.0 and NCCL 2.10"
            )
            mixed_precision = MixedPrecision(
                param_dtype=torch.bfloat16,
                reduce_dtype=torch.bfloat16,
                buffer_dtype=torch.bfloat16,
            )

        self.accelerator = accelerate.Accelerator(
            # See note in TransformerHandler._setup_accelerator() about this
            # mixed_precision=("bf16" if bf16 else "no"),
            fsdp_plugin=accelerate.FullyShardedDataParallelPlugin(
                sharding_strategy=ShardingStrategy.FULL_SHARD,
                mixed_precision_policy=mixed_precision,
                auto_wrap_policy="transformer_based_wrap",
                cpu_offload=self.worker_config.offload_cpu,
                activation_checkpointing=self.worker_config.activation_checkpointing,
                cpu_ram_efficient_loading=self.worker_config.cpu_ram_efficient_loading,
                sync_module_states=self.worker_config.cpu_ram_efficient_loading,
                state_dict_type=self.worker_config.state_dict_type,
                backward_prefetch=self.worker_config.backward_prefetch,
            ),
        )

        if self.config.lm_config.device == "meta":
            self.module = prepare_model_for_fsdp_with_meta_device(self.module)

        self.module = self.accelerator.prepare(self.module)

    def set_seed(self, seed: int) -> None:
        """Set the seed for the current worker."""
        set_seed(seed)

    def _exec_func(
        self,
        func: Callable[Concatenate[Self, TParams], TReturn] | str,
        *args,
        **kwargs,
    ) -> TReturn:
        # data will be on CPU when sent from controller
        data_device = _get_data_device()
        to_device = partial(_move_tensor, device=data_device)
        args = tree.map_structure(to_device, args)
        kwargs = tree.map_structure(to_device, kwargs)

        try:
            device_type: str = str(self.module.device.type)
            assert self.module.dtype is not None
            dtype: torch.dtype = self.module.dtype  # type: ignore[assignment]
            with torch.autocast(device_type=device_type, dtype=dtype):
                res = (
                    getattr(self, func)(*args, **kwargs)
                    if isinstance(func, str)
                    else func(self, *args, **kwargs)
                )

            # Needed to prevent GPU memory leak to the main process scheduling the workers
            if isinstance(res, GenerateDecoderOnlyOutput):
                res.past_key_values = None
                res["past_key_values"] = None

            to_cpu = partial(_move_tensor, device=torch.device("cpu"))
            return tree.map_structure(to_cpu, res)
        except Exception as e:
            # Re-raise the exception with traceback preserved. For some exceptions, Dask
            # modifies or loses the original traceback when crossing process boundaries.
            # RuntimeError preserves the traceback when using with_traceback() of original
            # exception.
            raise RuntimeError(str(e)).with_traceback(e.__traceback__)  # noqa: B904

    def __del__(self) -> None:
        dist.destroy_process_group()


class ParallelAsyncTransformer(AsyncTransformerInterface):
    def __init__(self, config: TransformerHandlerConfig):
        self._initialized = False

        parallel_mode_config = config.parallel_mode_config
        if not parallel_mode_config:
            raise ValueError("Parallel mode config must be provided.")
        self.config = config
        self.tokenizer = config.lm_config.get_tokenizer()
        maybe_set_tokenizer_chat_template(
            self.tokenizer, self.config.lm_config.chat_template
        )

        match parallel_mode_config.execution_mode:
            # TODO: see if we can just access `parallel_mode_config` as a
            # `config` attribute instead of passing both.
            case ExecutionMode.LOCAL_MACHINE:
                self._init_local_cluster(config, parallel_mode_config)
            case ExecutionMode.SLURM_CLUSTER:
                self._init_slurm_cluster(config, parallel_mode_config)
            case _:
                assert_never(parallel_mode_config.execution_mode)

        self._initialized = True

        atexit.register(self.teardown)

        # don't call AsyncTorchModule.__init__ because we don't need to set up module[_call_fn]
        AsyncBufferedWorker.__init__(
            self,
            batch_size=config.batch_size,
            max_wait_interval=config.max_wait_interval,
            collate_fn=config.collate_fn,
            decollate_fn=config.decollate_fn,
        )

        def handler_call_fn(handler: ParallelTransformerHandler, *args, **kwargs):
            return config.module_call_fn(handler.module, *args, **kwargs)

        self.handler_call_fn = handler_call_fn

    def _init_local_cluster(
        self, config: TransformerHandlerConfig, parallel_mode_config: ParallelModeConfig
    ):
        """Initialize a Dask cluster on local machine."""
        # lazy import since dask-cuda only works on Linux machines
        from dask_cuda import LocalCUDACluster

        # This uses NVIDIA's NVML layer instead of native CUDA, which is more robust in GPU detection
        # post initialization. This prevents issues with forked processes wrongly detecting the
        # default GPU as cuda:0
        os.environ["PYTORCH_NVML_BASED_CUDA_CHECK"] = "1"
        self.cluster = LocalCUDACluster(
            n_workers=parallel_mode_config.num_workers,
            threads_per_worker=parallel_mode_config.num_cpus_per_worker,
            host=parallel_mode_config.scheduler_addr,
            port=parallel_mode_config.scheduler_port,
            memory_limit=None,  # do not let Dask manage memory - if we OOM, we OOM
        )
        self._initialize_workers(config, parallel_mode_config)

    def _init_slurm_cluster(
        self, config: TransformerHandlerConfig, parallel_mode_config: ParallelModeConfig
    ):
        """Initialize a SLURM-based Dask cluster with GPU allocation."""
        # Lazy import because dask_jobqueue cannot be started in a subprocess, which
        # happens e.g. with streamlit
        from dask_jobqueue import SLURMCluster

        self.cluster = SLURMCluster(
            cores=parallel_mode_config.num_cpus_per_worker,
            memory=parallel_mode_config.memory,
            processes=1,  # Single dask worker per slurm worker
            walltime=parallel_mode_config.walltime,
            job_extra=[
                "--gres=gpu:1"
            ],  # 1 GPU per worker seems to be the common case for now
            log_directory=parallel_mode_config.log_directory,
        )
        self._initialize_workers(config, parallel_mode_config)

    def _initialize_workers(
        self, config: TransformerHandlerConfig, parallel_mode_config: ParallelModeConfig
    ):
        self.cluster.scale(parallel_mode_config.num_workers)
        self.client = Client(self.cluster)
        self.client.wait_for_workers(parallel_mode_config.num_workers)

        def get_cuda_visible_devices() -> int | None:
            device = os.environ.get("CUDA_VISIBLE_DEVICES", None)
            if device is not None:
                # If has several devices, assume the first one is the one to use for that worker
                if "," in device:
                    device = device.split(",", maxsplit=1)[0]
                    os.environ["CUDA_VISIBLE_DEVICES"] = device
                    os.environ["CUDA_VISIBLE_DEVICES"] = device
                return int(device)
            return None

        worker_to_cuda_device = self.client.run(get_cuda_visible_devices)
        workers_info = self.client.scheduler_info()["workers"]
        sorted_workers = dict(
            sorted(workers_info.items(), key=lambda item: item[1]["id"])
        )
        # The first worker is the master in the torch distributed setup
        master_addr = next(iter(sorted_workers.values()))["host"]

        futures = []
        worker_ids = []
        for rank, (worker_address, worker_data) in enumerate(sorted_workers.items()):
            worker_id = worker_data["id"]
            worker_cuda_device = worker_to_cuda_device[worker_address]
            if worker_cuda_device is None:
                assert (
                    parallel_mode_config.execution_mode != ExecutionMode.SLURM_CLUSTER
                ), "CUDA_VISIBLE_DEVICES should be pre set for SLURM workers."
                worker_cuda_device = rank

            parallel_worker_config = ParallelWorkerConfig(
                rank=rank,
                world_size=parallel_mode_config.num_workers,
                local_rank=worker_cuda_device,
                master_addr=master_addr,
                master_port=parallel_mode_config.torch_port,
                **parallel_mode_config.model_dump(),
            )
            future_op = self.client.submit(
                ParallelTransformerHandler,
                config=config,
                parallel_worker_config=parallel_worker_config,
                workers=[worker_id],
                actor=True,
            )
            futures.append(future_op)
            worker_ids.append(worker_id)

        self.actors: list[Actor] = self._client_gather(futures)
        self.worker_ids = worker_ids

    async def __call__(
        self,
        inputs: str | BatchEncoding | dict | list[dict] | None = None,
        tools_json: list[dict[Any, Any]] | None = None,
        input_ids: torch.Tensor | None = None,
        attention_mask: torch.Tensor | None = None,
        **kwargs,
    ) -> tuple[str, torch.Tensor]:
        if inputs is None:
            inputs = {"input_ids": input_ids, "attention_mask": attention_mask}
        inputs_tokenized = _get_tokenized_inputs(self.tokenizer, inputs, tools_json)
        inputs_len = inputs_tokenized["input_ids"].shape[1]
        outputs = await AsyncBufferedWorker.__call__(self, **inputs_tokenized, **kwargs)
        AsyncTransformer._maybe_finalize_logits_processors(
            kwargs.get("logits_processor"), outputs
        )

        return _process_outputs(self.config, self.tokenizer, outputs, inputs_len)

    async def _batched_call(self, batch_kwargs: dict[str, Any]):
        return self._submit_and_gather(
            self.handler_call_fn, **batch_kwargs, split_data=True
        )

    @overload
    def _submit_and_gather(
        self,
        func: Callable[Concatenate[ParallelTransformerHandler, TParams], TOutputType]
        | str,
        *args,
        split_data: Literal[True],
        **kwargs,
    ) -> TOutputType: ...

    @overload
    def _submit_and_gather(
        self,
        func: Callable[Concatenate[ParallelTransformerHandler, TParams], TReturn] | str,
        *args,
        split_data: Literal[False] = False,  # default
        **kwargs,
    ) -> list[TReturn]: ...

    def _submit_and_gather(
        self,
        func: Callable[Concatenate[ParallelTransformerHandler, TParams], Any] | str,
        *args,
        split_data: bool = False,
        **kwargs,
    ) -> TOutputType | list[TReturn]:
        """Submit a function to all workers and gather the results.

        Args:
            func: The function to send to each worker. If a string is provided,
                then getattr(handler, func) is used. If func is not a string,
                the first argument must be the ParallelTransformerHandler that it will
                be executed on.
            split_data: If True, split the data between workers. If False,
                send the same data to all workers.
            *args: Positional arguments to pass to the method.
            **kwargs: Keyword arguments to pass to the method.

        Returns:
            The gathered results from the workers.
        """
        if split_data:
            chunker = TensorChunker(
                num_chunks=len(self.actors),
            )
            split_args, split_kwargs, dummy_flags = chunker.chunkify(*args, **kwargs)
        else:
            split_args = [args] * len(self.actors)
            split_kwargs = [kwargs] * len(self.actors)

        futures = [
            handler._exec_func(
                func,
                *args_i,
                **kwargs_i,
            )
            for handler, worker_id, args_i, kwargs_i in zip(
                self.actors, self.worker_ids, split_args, split_kwargs, strict=True
            )
        ]
        results = self._client_gather(futures)

        if split_data:
            return chunker.dechunkify(results, dummy_flags)
        return results

    def wrap_afunc(
        self,
        func: Callable[
            Concatenate[ParallelTransformerHandler, TParams], Awaitable[TReturn]
        ],
        **kwargs,
    ) -> Callable[TParams, Awaitable[TReturn]]:
        raise NotImplementedError(
            "ParallelAsyncTransformer does not implement wrap_afunc(). "
            "Wrap a synchronous function with wrap_func() instead."
        )

    @overload
    def wrap_func(
        self,
        *,
        worker_agg_fn: Callable[[list[TReturn]], TReturn] | None = None,
        **kwargs,
    ) -> Callable[
        [Callable[Concatenate[ParallelTransformerHandler, TParams], TReturn]],
        Callable[TParams, TReturn],
    ]: ...

    @overload
    def wrap_func(
        self,
        func: Callable[Concatenate[ParallelTransformerHandler, TParams], TReturn],
        *,
        worker_agg_fn: Callable[[list[TReturn]], TReturn] | None = None,
        **kwargs,
    ) -> Callable[TParams, TReturn]: ...

    def wrap_func(
        self,
        func: (
            Callable[Concatenate[ParallelTransformerHandler, TParams], TReturn] | None
        ) = None,
        *,
        worker_agg_fn: Callable[[list[TReturn]], TReturn] | None = None,
        **kwargs,
    ) -> Callable:
        """Wrap a function to execute on all workers and return gathered results.

        Args:
            func: The function to wrap.
            worker_agg_fn: A function to aggregate the results from all workers.
            kwargs: Arguments that are discarded. Included here to enable a
                subclass to add additional arguments.
        """
        if worker_agg_fn is None:
            raise ValueError("worker_agg_fn must be provided.")

        if func is None:
            return partial(self.wrap_func, worker_agg_fn=worker_agg_fn, **kwargs)

        @wraps(func)
        def wrapped_func(*args, **kwargs) -> TReturn:
            return worker_agg_fn(self._submit_and_gather(func, *args, **kwargs))

        return wrapped_func

    def state_dict(self, **kwargs) -> dict[str, torch.Tensor]:
        # do this manually - accelerator.get_state_dict doesn't expose all options
        def state_dict_worker(
            handler: ParallelTransformerHandler, **kwargs
        ) -> dict[str, torch.Tensor]:
            # For an explanation, see:
            # https://pytorch.org/docs/stable/fsdp.html#torch.distributed.fsdp.FullStateDictConfig
            cfg = FullStateDictConfig(
                **({"offload_to_cpu": True, "rank0_only": True} | kwargs)
            )
            with FullyShardedDataParallel.state_dict_type(
                handler.module, StateDictType.FULL_STATE_DICT, cfg
            ):
                # will be the full state on rank 0 and empty on the others
                return handler.module.state_dict()

        # Only the 0th rank returns the full state dict
        state_dict = self._submit_and_gather(state_dict_worker, **kwargs)[0]
        return {k: v.cpu() for k, v in state_dict.items()}

    def load_state_dict(self, state_dict: dict[str, torch.Tensor]) -> None:
        # For some reason, Dask hangs when we pass a large object (e.g. state_dict)
        # directly to the workers. I can replicate it with the following:
        #
        # @handler.wrap_func
        # def hello(handler, _):
        #     print("hello")
        #
        # hello([0] *  1_000_000)
        # NOTE: this does not seem to be FSDP-related, as the issue didn't go away when
        # I disabled FSDP.
        raise NotImplementedError(
            "ParallelAsyncTransformer.load_state_dict() is not implemented yet. It is"
            " recommended to use .save_checkpoint() and .load_checkpoint() instead. "
        )

    def load_checkpoint(self, ckpt: os.PathLike | str) -> None:
        self._submit_and_gather("load_checkpoint", ckpt)

    def save_checkpoint(self, ckpt: os.PathLike | str, **kwargs) -> None:
        self._submit_and_gather("save_checkpoint", ckpt, **kwargs)

    def teardown(self) -> None:
        if self._initialized:
            self.client.shutdown()
            self.cluster.close()
            del self.client
            del self.cluster
            self._initialized = False

    def __del__(self) -> None:
        self.teardown()

    @staticmethod
    def _wrap_dask_future(dask_future: ActorFuture):
        """Converts a Dask ActorFuture into an awaitable asyncio.Future."""
        loop = asyncio.get_running_loop()
        return asyncio.ensure_future(loop.run_in_executor(None, dask_future.result))

    @staticmethod
    def _raise_exceptions(done, pending, wrapped_futures):
        exceptions = []
        for future in done:
            exc = future.exception()
            if exc:
                exceptions.append(exc)
        if exceptions:
            if len(exceptions) == 1:
                raise exceptions[0]
            raise ExceptionGroup("Multiple actor exceptions", exceptions)

        if pending:
            pending_indices = sorted([wrapped_futures.index(p) for p in pending])
            raise TimeoutError(
                f"Tasks didn't complete within timeout. {len(pending)} out of {len(wrapped_futures)} "
                f"still pending. Pending task indices: {pending_indices}"
            )

    async def _client_gather_async(self, futures):
        """Gather results from futures, propagating exceptions as they arrive.

        Unlike client.gather() which waits for all futures to complete before raising
        any exceptions, this method processes futures as they complete and raises
        exceptions immediately. This is crucial when using FSDP where workers may
        be stuck waiting for each other when one worker crashes, causing long hangs.

        Note: Dask Actors currently have an issue where they're not working properly with
        dask.gather() and can cause blocking issues or hide worker errors. This implementation
        works around those limitations.
        """
        try:
            wrapped_futures = [self._wrap_dask_future(f) for f in futures]

            # Use asyncio.wait with FIRST_EXCEPTION instead of gather
            done, pending = await asyncio.wait(
                wrapped_futures, timeout=1200, return_when=asyncio.FIRST_EXCEPTION
            )

            self._raise_exceptions(done, pending, wrapped_futures)

            return await asyncio.gather(*wrapped_futures)
        except Exception:
            logger.exception("Error in dask workers: %s")
            for future in wrapped_futures:
                future.cancel()
            self.teardown()
            # sys.exit(1) would wait for dask to finish, which can cause hanging
            # when workers are in a deadlock. Use os._exit to force immediate termination
            # TODO: this is more of a hack, we should propagate special exception that is
            # not caught by the rollout manager.
            os._exit(1)

    def _client_gather(self, futures: list[ActorFuture]) -> list[Any]:
        # Use distributed.utils.sync to run the async function in the current thread
        return sync(self.client.loop, self._client_gather_async, futures)  # type: ignore[arg-type]


# Helpers


# TODO: We should use the tokenizer to manage this instead of calling pad ourselves.
def _collate_fn_transformer(
    samples: list[dict[str, torch.Tensor]],
    pad_token_id: int,
    left_pad: bool,
    agg_keys: set[str] | None = None,
) -> dict[str, torch.Tensor]:
    """Collates and pads a batch of samples for input into a huggingface transformer model."""
    if agg_keys is None:
        agg_keys = {"input_ids", "attention_mask"}
    seq_lens = [inp["input_ids"].shape[1] for inp in samples]
    max_seq_len = max(seq_lens)
    n_pads = [max_seq_len - seq_len for seq_len in seq_lens]

    # TODO: these pad_value and pad side lookups might be getting expensive. Refactor
    pad_value = {
        "input_ids": pad_token_id,
        "attention_mask": 0,
    }

    batch = {
        key: torch.cat(
            [
                # see comment in TransformerHandler.__init__ about why we're left padding.
                pad(
                    inp[key],
                    ((n_pad, 0) if left_pad else (0, n_pad)),
                    value=pad_value.get(key, 0),
                )
                for inp, n_pad in zip(samples, n_pads, strict=True)
            ],
            dim=0,
        )
        for key in agg_keys
    }

    # Treating other keys as constant kwargs params for the model
    other_keys = set(samples[0].keys()) - agg_keys
    for key in other_keys:
        for sample in samples:
            if key not in sample:
                raise ValueError(f"Missing key {key} in sample.")
            if key in batch and batch[key] != sample[key]:
                raise ValueError(
                    f"Constant kwarg key {key} has different values within batch."
                )
            batch[key] = sample[key]

    return batch


# be explicit
collate_fn_transformer_left_pad = partial(_collate_fn_transformer, left_pad=True)
collate_fn_transformer_right_pad = partial(_collate_fn_transformer, left_pad=False)


def decollate_fn_transformer_decoder(
    batched_output: GenerateDecoderOnlyOutput,
) -> list[GenerateDecoderOnlyOutput]:
    """Decollates a batched output from a huggingface transformer decoder."""
    batch_size = len(batched_output.sequences)
    outputs: list[GenerateDecoderOnlyOutput] = [
        GenerateDecoderOnlyOutput(
            sequences=batched_output.sequences[i][None, :],  # type: ignore[arg-type]
            scores=[
                score[i][None, :]
                for score in batched_output.scores
                if (score[i] is not None)
            ]
            if batched_output.scores
            else None,  # type: ignore[arg-type]
        )
        for i in range(batch_size)
    ]
    # There are other fields in the batched output that we can add here,
    # but no calling code is using them, so ignore for now.
    return outputs


# From: https://github.com/pytorch/torchtune/blob/c5db813ce0473db090a4f1f6b450f559acac58e5/torchtune/training/_distributed.py#L207
# NOTE: I removed the LoRA logic from the source code since we don't use torchtune for now.
# See issue in docstring for why we needed this.
def prepare_model_for_fsdp_with_meta_device(model: nn.Module) -> nn.Module:
    """Dynamically define reset_parameters on every submodule of the model.

    More details here: https://github.com/pytorch/pytorch/issues/104187.

    Args:
        model: model to prepare for usage with FSDP and meta device.

    Returns:
        Model with reset_parameters defined on every submodule.

    Raises:
        RuntimeError: if model contains submodule with non-callable attribute reset_parameters
    """
    for k, v in model.named_modules():
        # If the module does not have reset_parameters defined, we define
        # a no-op reset_parameters method to satisfy FSDP's contract.
        reset_params = getattr(v, "reset_parameters", None)

        if reset_params is not None and not callable(reset_params):
            raise RuntimeError(
                "Cannot override existing reset_parameters variable for FSDP init"
                f" in {k}"
            )

        if reset_params is None:
            v.reset_parameters = _dummy_reset_params.__get__(v)  # pylint: disable=assignment-from-no-return

    return model


def _dummy_reset_params(_: nn.Module) -> None:
    """Dummy method for patching no-op reset_parameters() when using FSDP with meta device."""


@cache
def _get_data_device() -> torch.device:
    """Get the device where input tensors should be placed."""
    if torch.cuda.is_available():
        return torch.device("cuda")
    return torch.device("cpu")


def _get_tokenized_inputs(
    tokenizer: PreTrainedTokenizer,
    inputs: str | dict | BatchEncoding | list[dict],
    tools_json: list[dict[Any, Any]] | None = None,
) -> BatchEncoding:
    if isinstance(inputs, BatchEncoding):
        return inputs
    if isinstance(inputs, dict):
        return BatchEncoding(inputs)
    if isinstance(inputs, str):
        return tokenizer(inputs, return_tensors="pt")
    if is_conversation(inputs):
        result = tokenizer.apply_chat_template(
            inputs,
            tools=tools_json,  # type: ignore[arg-type]
            return_tensors="pt",
            return_dict=True,
            add_generation_prompt=True,
        )
        return cast(BatchEncoding, result)
    raise ValueError(
        f"inputs must be a str, BatchEncoding, or Conversation, but got {type(inputs)}"
    )


def _process_outputs(
    config: TransformerHandlerConfig,
    tokenizer: PreTrainedTokenizer,
    outputs,
    input_len: int,
) -> tuple[str, torch.Tensor]:
    outputs = tree.map_structure(lambda x: x.cpu(), outputs)

    if config.lm_type == LMType.REGRESSION:
        # TODO: deal with different output types w @overload
        # dask cannot serialize precisions lower than fp32
        return "", outputs.float()

    assert len(outputs.sequences) == 1, (
        "Expected a batch size of 1, as AsyncBufferedWorker handles batching internally."
    )
    output_sequence = outputs.sequences[0]

    if outputs.scores:
        output_sequence = output_sequence[-len(outputs.scores) :]
    else:
        # This way of finding the output sequence can be messed up by tokenizers that
        # have overlapping eos/bos/pad tokens. We should consider removing it and relying
        # on `outputs.scores` exclusively.
        def _find_first_non_padding_index(
            sequence: torch.Tensor, pad_token_id: int
        ) -> int:
            """Find the first non-padding index in a sequence."""
            padding_mask = (sequence == pad_token_id).long()
            cumsum_padding = padding_mask.cumsum(dim=0)  # noqa: FURB184
            sequence_length = sequence.size(0)
            expected_cumsum = torch.arange(1, sequence_length + 1)
            divergence = (cumsum_padding != expected_cumsum).long()
            # Find the first index where the cumsums diverge. This is the first non-padding index.
            return int(divergence.argmax(dim=0).item())

        padding_len = _find_first_non_padding_index(
            output_sequence, tokenizer.pad_token_id
        )
        # Take the output sequence starting after the initial padding and input length
        output_sequence = output_sequence[padding_len + input_len :]

    out_text = tokenizer.decode(output_sequence, skip_special_tokens=True)

    # Extract logprobs
    if outputs.scores:
        logits = torch.cat(outputs.scores, dim=0).to(torch.float32)
        if len(logits) != len(output_sequence):
            discarded_logits = logits[len(output_sequence) :]
            assert torch.all(discarded_logits == -float("inf")), (
                "Discarded logits should be invalid ones."
            )
            # Discard the logits for post-eos tokens
            logits = logits[: len(output_sequence)]
        logprobs = logits_to_logprobs(logits, output_sequence)
    else:
        logprobs = torch.tensor([])

    return out_text, logprobs


def logits_to_logprobs(logits: torch.Tensor, token_ids: torch.Tensor) -> torch.Tensor:
    """Convert logits to log probabilities.

    Roughly following https://huggingface.co/Pwicke/logprobs_for_CausalLMs

    Note that PretrainedModel.generate returns a tuple of logits, but
    TransformerHandler._process_outputs converts it into tensor for us.
    """
    logprobs = nn.functional.log_softmax(logits, dim=-1)
    # Can also be written as logprobs[torch.arange(len(token_ids)), token_ids]
    # but gather is more efficient.
    return torch.gather(logprobs, dim=-1, index=token_ids.unsqueeze(-1)).squeeze(-1)


def _move_tensor(x, device: torch.device) -> torch.Tensor:
    return x.to(device) if isinstance(x, torch.Tensor) else x


def maybe_set_tokenizer_chat_template(
    tokenizer: PreTrainedTokenizer, chat_template: str | None
) -> None:
    """Set the chat template for the tokenizer if needed."""
    # Check if the tokenizer's existing chat_template contains the '{% generation %}' tag

    model_name = tokenizer.name_or_path.lower()
    if in_test := ("PYTEST_CURRENT_TEST" in os.environ and "gpt2" in model_name):
        # Override whatever was passed in for tests
        logger.info("Using llama3 chat template for gpt2 model in tests.")
        chat_template = "llama3_chat_template_ori.jinja"

    if chat_template is not None:
        # User should provided a chat template that includes the {% generation %} keyword needed
        # for populating "assistant_masks" when tokenizing inputs for the purpose of training.
        # We need to monitor the PR comments at https://github.com/huggingface/transformers/pull/30650
        # for updates on the issue.
        template_path = Path(chat_template)
        if not template_path.exists():
            # Check the aviary_internal library of templates
            template_path = Path(f"{REPO_ROOT}/ldp/nn/chat_templates/{chat_template}")
        loaded_chat_template = template_path.read_text()

        if in_test:
            # Adapt the template for the actual tokenizer we are using
            loaded_chat_template = loaded_chat_template.replace(
                "<|eot_id|>", tokenizer.eos_token
            )
        tokenizer.chat_template = loaded_chat_template
        return

    if not tokenizer.chat_template or "{% generation %}" in tokenizer.chat_template:
        # Warn the user about potential training issues
        logger.warning(
            "Tokenizer does not have a chat template with '{% generation %}'."
            " Generative training will have issues as the tag does not exist, which"
            " means HuggingFace's internal code for retrieving the assistant_mask for"
            " training will be invalid. Fine-tuning for other purposes, like"
            " regression, will not be affected."
        )



import asyncio
import collections
import inspect
import itertools
import random
from collections.abc import Awaitable, Callable, Hashable, Iterable, Sequence
from typing import TYPE_CHECKING, Any, Literal, TypeVar

if TYPE_CHECKING:
    import networkx as nx
import numpy as np
from aviary.core import Message, Tool, ToolRequestMessage, join

from ldp.graph import OpResult
from ldp.graph.op_utils import _lazy_import_networkx
from ldp.graph.ops import GradOutType


def to_network(  # noqa: C901
    op_result: OpResult,
    max_label_height: int | None = None,
    max_label_width: int | None = None,
    G: "nx.MultiDiGraph | None" = None,
) -> "nx.MultiDiGraph":
    """
    Populate a NetworkX graph from the input op result's computation graph.

    How to export Graphviz .dot file: nx.drawing.nx_pydot.write_dot(G, "file.dot")
    How to render with Graphviz: nx.drawing.nx_pydot.to_pydot(G).write_png("file.png")
    Online Graphviz renderer: https://dreampuf.github.io/GraphvizOnline/

    Args:
        op_result: Starting op result to recurse parent op calls and results.
        max_label_height: Optional max label height (lines).
        max_label_width: Optional max label width (chars).
        G: Optional graph to add nodes/edges to. Allows this to be a recursive function.

    Returns:
        Populated a NetworkX multi-edge directed graph.
    """
    nx = _lazy_import_networkx()

    def gvizify(x: Any) -> str:
        """Stringify and then escape colons for Graphviz labels."""
        if isinstance(x, OpResult):
            x = x.value
        if isinstance(x, Sequence):
            if isinstance(x[0], Message):
                x = join(x)
            elif isinstance(x[0], Tool):
                x = "\n".join(f"Tool {t.info.name}" for t in x)
        elif isinstance(x, ToolRequestMessage):
            # reformatting tool calls to make them easier to read
            x = str(x).split(" for tool calls: ")
            x = "\n".join(x).replace("; ", "\n")
        result = (
            "\n".join(
                # Replace double quotes since they can interfere with colon escapes
                # Strip here to avoid trailing spaces in the labels
                x_line[:max_label_width].replace('"', "'").strip()
                for i, x_line in enumerate(str(x).split("\n"))
                if not max_label_height or i < max_label_height
            )
        ).strip()  # Remove trailing newlines
        return result if ":" not in result else f'"{result}"'  # Escape colons

    call_id = op_result.call_id
    assert call_id is not None, (
        "to_network currently assumes a compute graph is available"
    )
    ctx = op_result.ctx

    op_result_str = gvizify(op_result)
    op_result_node = gvizify(f"{op_result_str}\n{call_id.fwd_id}")
    if G is None:
        # TODO: figure out a way to use OpResult.get_compute_graph(), which builds
        # a nx.DiGraph.
        G = nx.MultiDiGraph()

    op_call_str = gvizify(f"{ctx.op_name}:{call_id.fwd_id}")
    if op_call_str in G:
        # We have already visited this node - can skip.
        return G

    G.add_node(op_result_node, style="dotted", label=op_result_str)
    G.add_edge(op_call_str, op_result_node)

    if (
        result_grad := ctx.get(key="grad_output", call_id=call_id, default=None)
    ) is not None:
        G.add_edge(
            op_result_node,
            op_call_str,
            label=gvizify(result_grad),
            style="dotted",
        )

    input_args, input_kwargs = op_result.inputs
    grads = ctx.get(key="grad_input", call_id=call_id, default=None)
    if grads is None:
        arg_grads: list[GradOutType | None] = [None] * len(input_args)
        kwarg_grads: dict[str, GradOutType | None] = dict.fromkeys(input_kwargs)
    else:
        arg_grads, kwarg_grads = grads

    args_and_grads = itertools.chain(
        zip(input_args, arg_grads, strict=True),
        ((arg, kwarg_grads[key]) for key, arg in input_kwargs.items()),
    )

    for arg, grad in args_and_grads:
        arg_str = gvizify(arg)

        if not isinstance(arg, OpResult):
            G.add_node(arg_str, style="dotted")

        else:
            arg_str = gvizify(f"{arg_str}\n{arg.call_id.fwd_id}")
            G = to_network(
                arg,
                max_label_height=max_label_height,
                max_label_width=max_label_width,
                G=G,
            )

        G.add_edge(arg_str, op_call_str)
        if grad is not None:
            G.add_edge(op_call_str, arg_str, label=gvizify(grad), style="dotted")

    return G


def measure_consensus_proportion(
    consensus_size: int, sample_size: int
) -> tuple[float, float]:
    """Aggregate sampling accuracy mean and standard error on consensus proportion."""
    sample_acc_mean = consensus_size / sample_size
    # Use binomial standard error since we're comparing ideal vs not ideal
    sample_acc_ste = np.sqrt(sample_acc_mean * (1.0 - sample_acc_mean) / sample_size)
    return sample_acc_mean, sample_acc_ste


TData = TypeVar("TData")
TGroupKey = TypeVar("TGroupKey", bound=Hashable)
TAnswer = TypeVar("TAnswer")
NO_IDEAL_ANSWER_FN: Literal["NO_IDEAL_ANSWER_FN"] = "NO_IDEAL_ANSWER_FN"  # Sentinel


async def bulk_evaluate_consensus(
    data: Iterable[TData],
    grouping_fn: Callable[[TData], TGroupKey],
    extract_answer_fn: Callable[[TData], TAnswer | Awaitable[TAnswer]],
    num_samples: int | None = None,
    seed: np.random.Generator | random.Random | int | None = None,
    ideal_answer_fn: (
        Callable[[TData], TAnswer] | Literal["NO_IDEAL_ANSWER_FN"]
    ) = NO_IDEAL_ANSWER_FN,
    consensus_callback: Callable[[TAnswer, int, int], Any] | None = None,
) -> tuple[dict[TGroupKey, list[tuple[TAnswer, int]]], float]:
    """
    Create consensus groups and evaluate the consensus accuracy for each.

    Args:
        data: Flattened data to evaluate consensus upon. Think of this as all results
            from at least one evaluation upon a TaskDataset.
        grouping_fn: Function to extract the group key from a datum. For a QA dataset,
            the group key could be the question or question ID.
        extract_answer_fn: Passed through to evaluate_consensus.
        num_samples: Passed through to evaluate_consensus.
        seed: Passed through to evaluate_consensus.
        ideal_answer_fn: Optional function to extract the ideal answer from a datum to
            compute accuracy with, or pass NO_IDEAL_ANSWER to skip this calculation.
        consensus_callback: Passed through to evaluate_consensus.

    Returns:
        Two-tuple of (1) a dictionary mapping group keys to consensus list
            (SEE evaluate_consensus.__doc__'s Returns for more details), and
            (2) the proportion of groups for which the consensus matches the ideal.
    """
    groups = collections.defaultdict(list)
    for x in data:
        groups[grouping_fn(x)].append(x)

    grouped_consensus: dict[TGroupKey, list[tuple[TAnswer, int]]] = {}

    async def add_consensus_check_ideal(
        group_key: TGroupKey, group: list[TData]
    ) -> int:
        grouped_consensus[group_key], consensus = await evaluate_consensus(
            group, extract_answer_fn, num_samples, seed, consensus_callback
        )
        if ideal_answer_fn != NO_IDEAL_ANSWER_FN:  # If we have an ideal
            # Assume all items in the group have the same ideal answer
            return consensus == ideal_answer_fn(group[0])
        return 0

    ideal_count = sum(
        await asyncio.gather(
            *itertools.starmap(add_consensus_check_ideal, groups.items())
        )
    )
    return grouped_consensus, ideal_count / len(groups) if groups else 0.0


async def evaluate_consensus(
    data: Sequence[TData],
    extract_answer_fn: Callable[[TData], TAnswer | Awaitable[TAnswer]],
    num_samples: int | None = None,
    seed: np.random.Generator | random.Random | int | None = None,
    consensus_callback: Callable[[TAnswer, int, int], Any] | None = None,
) -> tuple[list[tuple[TAnswer, int]], TAnswer]:
    """
    Create consensus bins given data.

    Args:
        data: Data to evaluate consensus upon, length is called N.
        extract_answer_fn: Function to extract the actual answer from a datum. It can
            be async so this can be done using a LLM call.
        num_samples: Number of samples to choose from the N total, or None (default) to
            infer this value to match N.
        seed: Optional seed for sampling.
        consensus_callback: Optional callback function called just after computing
            consensus, it's passed the consensus answer, consensus size, and
            sample size. This is useful for stuff like summary statistics.

    Returns:
        Two-tuple containing (1) an ordered list of `(extracted_answer, count)` tuples,
            ordered by highest `count` first, and (2) the consensus answer
            (i.e. the `extracted_answer` with the highest `count`).
    """
    rand = (
        seed
        if isinstance(seed, np.random.Generator | random.Random)
        else np.random.default_rng(seed)
    )
    if num_samples is None:
        num_samples = len(data)
    if len(data) < num_samples:  # Too few items, sample with replacement
        raise ValueError(
            f"A number of samples {num_samples} exceeding the {len(data)} data points"
            " present is disallowed since sampling with replacement can produce"
            " misleading consensus. Imagine if there was 1 data point, but 100 samples,"
            " this would report perfect consensus with low standard error (but it would"
            " be statistically artificial)."
        )
    if isinstance(rand, random.Random):  # Built-in random sample without replacement
        sampled: Iterable[TData] = rand.sample(data, num_samples)
    else:  # NumPy random sample without replacement
        sampled = rand.choice(data, num_samples, replace=False)  # type: ignore[arg-type]

    async def extract_answer(datum: TData) -> TAnswer:
        answer = extract_answer_fn(datum)
        return await answer if inspect.isawaitable(answer) else answer

    # Get answers for the sampled data
    answers = await asyncio.gather(*(extract_answer(x) for x in sampled))
    # Compute consensus: mode of the sampled answers
    most_common = collections.Counter(answers).most_common()
    # NOTE: If there are multiple modes, just use the first one
    consensus_answer, consensus_count = most_common[0]
    if consensus_callback:
        consensus_callback(consensus_answer, consensus_count, num_samples)
    return most_common, consensus_answer


def compute_pass_at_k(n: int, c: int, k: int) -> float | np.floating:
    """Compute an unbiased estimation for 'pass @ k'.

    Source: https://doi.org/10.48550/arXiv.2107.03374's figure 3.

    If there's multiple tasks, an aggregation used is averaging pass @ k across tasks,
    per https://doi.org/10.48550/arXiv.2407.21787's equation 1.

    Args:
        n: Total number of samples.
        c: Number of correct (pass a verifier) samples.
        k: k term (number of attempts) used in pass @ k.

    Returns:
        Unbiased estimation for pass @ k, the probability of getting at least one
            successful outcome in k attempts.
    """
    if n - c < k:
        return 1.0
    return 1.0 - np.prod(1.0 - k / np.arange(n - c + 1, n + 1))



import asyncio
import copy
import math
from collections.abc import Callable, Iterable
from functools import partial
from typing import Any, cast

import numpy as np
import pytest
import torch
import tree

from ldp.graph import CallID, ConfigOp, FxnOp, Op, OpCtx, OpResult, compute_graph
from ldp.graph.gradient_estimators import (
    TorchParamBackwardEstimator,
    assign_constant_grads,
    assign_default_grads,
)
from ldp.graph.ops import GradInType, ResultOrValue
from ldp.graph.torch_ops import TorchOp


class PoissonSamplerOp(Op):
    @staticmethod
    def _probability(lam: float, k: int) -> float:
        if k < 0:
            # negative k can happen when taking a gradient
            return 0.0
        return np.exp(-lam) * lam**k / math.factorial(k)

    async def forward(self, lam: float) -> int:
        return np.random.poisson(max(0.01, lam))  # noqa: NPY002

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args,
        input_kwargs,
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        # This op has no internal parameters, so we just compute delta_{j,i} in section 4.
        lam = max(0.01, input_kwargs["lam"])
        k = ctx.get(call_id, "output").value

        p_k = cls._probability(lam, k)
        p_km1 = cls._probability(lam, k - 1)

        # dp(k)/dlam
        grad_lam_p = p_km1 - p_k

        # d[lnp(k)]/dlam
        grad_lam_lnp = grad_lam_p / p_k

        # define dk/dlam in expectation: dE[k]/dlam = dlam/dlam = 1
        grad_lam_k = 1.0

        # delta_{j,i}
        delta_lam = grad_lam_lnp + (grad_lam_k * cast("float", grad_output))

        return [], {"lam": delta_lam}


class FloatParamOp(Op):
    def __init__(self, init_param: float):
        self.param = init_param

    async def forward(self) -> float:
        return self.param

    @classmethod
    def backward(
        cls, ctx: OpCtx, input_args, input_kwargs, grad_output: Any, call_id: CallID
    ) -> GradInType:
        return [], {}


class SGDOptimizer:
    def __init__(self, op: FloatParamOp, lr: float = 0.01, lr_decay: float = 1.0):
        self.op = op
        self.lr = lr
        self.lr_decay = lr_decay

        self.accumulated_updates: list[float] = []

    def aggregate(self, samples: Iterable[tuple[OpResult, float]]):
        for result, reward in samples:
            assert result.call_id is not None
            call_ids = self.op.get_call_ids({result.call_id.run_id})

            # These are delta_{k,j} in equation 20
            grads = [
                cast("float", g)
                for g in (
                    self.op.ctx.get(call_id, "grad_output") for call_id in call_ids
                )
                if g is not None
            ]
            if not grads:
                # this op call was pruned from the backwards call graph
                continue

            # We assume FloatParamOp is deterministic, meaning grad_j lnp_j = 0
            # Furthermore, grad_j x_j = 1 (dlam/dlam), so Eq 20 reduces to R*sum_k(delta_{k,j})
            self.accumulated_updates.append(reward * sum(grads))

    def update(self):
        self.op.param += self.lr * cast("float", np.mean(self.accumulated_updates))
        self.accumulated_updates.clear()
        self.lr *= self.lr_decay


@pytest.mark.parametrize("init_lam", [0.1, 20.0, 10.0])
@pytest.mark.flaky(reruns=3, only_on=[AssertionError])
@pytest.mark.asyncio
async def test_poisson_sgd(init_lam: float):
    # This test optimizes the rate parameter of a Poisson distribution such that
    # the expected value of the distribution is 10.
    target = 10

    def reward_fn(k: int) -> float:
        return -np.abs(k - target)

    lam = FloatParamOp(init_lam)
    poisson = PoissonSamplerOp()
    opt = SGDOptimizer(lam, lr=1.0, lr_decay=0.9)

    @compute_graph()
    async def fwd() -> OpResult[int]:
        return await poisson(await lam())

    bsz = 4
    n_epochs = 20
    for _ in range(n_epochs):
        samples = await asyncio.gather(*[fwd() for _ in range(bsz)])

        training_batch: list[tuple[OpResult, float]] = []
        for k in samples:
            k.compute_grads()
            reward = reward_fn(k.value)
            training_batch.append((k, reward))

        opt.aggregate(training_batch)
        opt.update()

        print(lam.param, opt.lr, np.mean([k.value for k in samples]))

    # noisy learning, so just make sure we went in the right direction
    assert np.isclose(lam.param, target, atol=4.0)


def assign_constant_grads_alter_inputs(
    _ctx: OpCtx,
    input_args,
    input_kwargs,
    _grad_output: int,
    _call_id: CallID,
    grad_val: float = 0.0,
    input_func: Callable = lambda x, y: (x, y),
) -> GradInType:
    input_args, input_kwargs = copy.deepcopy(input_args), copy.deepcopy(input_kwargs)
    input_args, input_kwargs = input_func(input_args, input_kwargs)
    return assign_constant_grads(input_args, input_kwargs, grad_val)


@pytest.mark.asyncio
async def test_nested_dict_kwargs_grad_aggregation_success():
    """Tests gradient aggregation across two ops that return tree.Structure objects."""
    cfg = {"a": {"b": [1, 2, 3], "c": 4}}
    config_op = ConfigOp(cfg)

    op1 = FxnOp[int](lambda _input_dict: 2)
    op1.set_name("op1")
    op2 = FxnOp[int](lambda _input_dict: 3)
    op2.set_name("op2")
    agg_op = FxnOp[int](lambda *args: sum(args))  # noqa: FURB111

    backward1 = partial(assign_constant_grads_alter_inputs, grad_val=1.0)
    backward2 = partial(assign_constant_grads_alter_inputs, grad_val=2.0)

    @compute_graph()
    async def fwd() -> OpResult[int]:
        config = await config_op()
        a = await op1(config)
        b = await op2(config)
        return await agg_op(a, b)

    output = await fwd()
    output.compute_grads(backward_fns={"op1": backward1, "op2": backward2})
    assert output.value == 5
    config_op_grad = output.inputs[0][0].inputs[1]["_input_dict"].grad
    assert config_op_grad == {"a": {"b": [3.0, 3.0, 3.0], "c": 3.0}}


@pytest.mark.asyncio
async def test_nested_dict_kwargs_missing_inner_grad_aggregation_fail():
    """Tests that two ops that return different tree.Structure objects crash in grad aggregation."""
    cfg = {"a": {"b": [1, 2, 3], "c": 4}}
    config_op = ConfigOp(cfg)

    op1 = FxnOp[int](lambda _input_dict: 3)
    op2_missing_grad = FxnOp[int](lambda _input_dict: 4)
    op2_missing_grad.set_name("op2_missing_grad")
    agg_op = FxnOp[int](lambda *args: sum(args))  # noqa: FURB111

    def backward_with_missing_leaf(input_args, input_kwargs):
        del input_kwargs["_input_dict"]["a"]["c"]  # missing gradient for this key
        return input_args, input_kwargs

    backward_with_missing_leaf = partial(
        assign_constant_grads_alter_inputs, input_func=backward_with_missing_leaf
    )

    @compute_graph()
    async def fwd() -> OpResult[int]:
        config = await config_op()
        a = await op1(config)
        b = await op2_missing_grad(config)
        return await agg_op(a, b)

    output = await fwd()
    with pytest.raises(ValueError, match="Mismatched gradient structures"):
        output.compute_grads(
            backward_fns={"op2_missing_grad": backward_with_missing_leaf}
        )


@pytest.mark.asyncio
async def test_nested_dict_kwargs_missing_or_extra_inner_grad_ok():
    """Tests that missing or extra gradients for a dict input are not failing."""
    input_dict = {"a": {"b": [1, 2, 3], "c": 4}}

    op = FxnOp[int](lambda _input_dict: 4)
    op.set_name("op")
    agg_op = FxnOp[int](lambda *args: sum(args))  # noqa: FURB111

    def backward_with_missing_leaf(input_args, input_kwargs):
        del input_kwargs["_input_dict"]["a"]["c"]  # missing gradient for this key
        return input_args, input_kwargs

    backward_with_missing_leaf = partial(
        assign_constant_grads_alter_inputs, input_func=backward_with_missing_leaf
    )

    @compute_graph()
    async def fwd() -> OpResult[int]:
        a = await op(input_dict)
        return await agg_op(a)

    output = await fwd()
    output.compute_grads(backward_fns={"op": backward_with_missing_leaf})
    assert output.value == 4

    def backward_with_extra_leaf(input_args, input_kwargs):
        input_kwargs["_input_dict"]["a"]["d"] = 3  # add extra key
        return input_args, input_kwargs

    backward_with_extra_leaf = partial(
        assign_constant_grads_alter_inputs, input_func=backward_with_extra_leaf
    )

    output = await fwd()
    output.compute_grads(backward_fns={"op": backward_with_extra_leaf})
    assert output.value == 4


@pytest.mark.asyncio
async def test_args_missing_or_extra_grad():
    op = FxnOp[int](lambda x: x)
    agg_op = FxnOp[int](lambda *args: sum(args))  # noqa: FURB111
    agg_op.set_name("agg_op")

    def backward_with_missing_arg(input_args, input_kwargs):
        input_args = input_args[:-1]  # remove last arg
        return input_args, input_kwargs

    backward_with_missing_arg = partial(
        assign_constant_grads_alter_inputs, input_func=backward_with_missing_arg
    )

    @compute_graph()
    async def fwd() -> OpResult[int]:
        a = await op(1)
        b = await op(2)
        return await agg_op(a, b)

    output = await fwd()
    with pytest.raises(ValueError, match="argument 2 is shorter than argument 1"):
        output.compute_grads(backward_fns={"agg_op": backward_with_missing_arg})

    def backward_with_extra_arg(input_args, input_kwargs):
        input_args += [3]  # add extra arg
        return input_args, input_kwargs

    backward_with_extra_arg = partial(
        assign_constant_grads_alter_inputs, input_func=backward_with_extra_arg
    )

    output = await fwd()
    with pytest.raises(ValueError, match="argument 2 is longer than argument 1"):
        output.compute_grads(backward_fns={"agg_op": backward_with_extra_arg})


@pytest.mark.asyncio
async def test_kwargs_missing_or_extra_grad() -> None:
    input1 = 1
    input2 = 2
    # Don't use FURB118 here because we assert on kwargs below
    op = FxnOp[int](lambda _input1, _input2: _input1 + _input2)  # noqa: FURB118
    op.set_name("op")

    def backward_with_missing_kwarg(input_args, input_kwargs):
        del input_kwargs["_input1"]
        return input_args, input_kwargs

    backward_with_missing_kwarg = partial(
        assign_constant_grads_alter_inputs, input_func=backward_with_missing_kwarg
    )

    @compute_graph()
    async def fwd() -> OpResult[int]:
        return await op(input1, input2)

    output = await fwd()
    with pytest.raises(ValueError, match="Mismatch between grads"):
        output.compute_grads(backward_fns={"op": backward_with_missing_kwarg})

    def backward_with_extra_kwarg(input_args, input_kwargs):
        input_kwargs["_input3"] = 3
        return input_args, input_kwargs

    backward_with_extra_kwarg = partial(
        assign_constant_grads_alter_inputs, input_func=backward_with_extra_kwarg
    )

    output = await fwd()
    with pytest.raises(ValueError, match="Mismatch between grads"):
        output.compute_grads(backward_fns={"op": backward_with_extra_kwarg})


@pytest.mark.asyncio
async def test_assign_default_grads():  # noqa: RUF029
    input_args: list[ResultOrValue] = [1, 2]
    input_kwargs: dict[str, ResultOrValue] = {
        "a": {"b": 3, "c": 4},
        "d": {"e": {"f": 5}},
    }
    input_grad_args = [0.1]
    input_grad_kwargs = {"a": {"b": 0.2}}

    expected_grad_args = [0.1, 7.0]
    expected_grad_kwargs = {
        "a": {
            "b": 0.2,
            "c": 7.0,
        },
        "d": {"e": {"f": 7.0}},
    }

    input_grads = (input_grad_args, input_grad_kwargs)
    output_grad_args, output_grad_kwargs = assign_default_grads(
        input_grads,
        input_args,
        input_kwargs,
        default_grad_val=7.0,
    )

    assert output_grad_args == expected_grad_args
    assert output_grad_kwargs == expected_grad_kwargs


# test running 2 ops serially without calling @compute_graph
@pytest.mark.asyncio
async def test_serial_ops_diff_run_id():
    op1 = FxnOp[int](lambda x: x + 1)
    op2 = FxnOp[int](lambda x: x * 2)

    result1 = await op1(1)

    with pytest.raises(RuntimeError, match="args and kwargs must have the same run_id"):
        await op2(result1)


@pytest.mark.parametrize("hidden_nodes", [1, 4])
@pytest.mark.asyncio
async def test_torch_param_backward_estimator(hidden_nodes: int):
    torch_module = torch.nn.Sequential(
        torch.nn.Linear(4, hidden_nodes),
        torch.nn.Linear(hidden_nodes, 1),
    )
    torch_op = TorchOp(torch_module)
    estimator = TorchParamBackwardEstimator(torch_module)

    # Forward pass
    result = await torch_op(torch.randn(4, requires_grad=True))

    # Backward pass
    result.compute_grads(backward_fns={torch_op.name: estimator.backward})

    # Check that the gradients are computed and have the correct shape
    call_ids = torch_op.get_call_ids({result.call_id.run_id})
    grad_params = torch_op.ctx.get(next(iter(call_ids)), "grad_params")

    for named_param, grad_param in torch_module.named_parameters():
        assert named_param in grad_params
        assert grad_param.shape == grad_params[named_param].shape



# Lower than LiteLLM's 10-min default: https://github.com/BerriAI/litellm/blob/v1.48.10/litellm/main.py#L859
DEFAULT_LLM_COMPLETION_TIMEOUT = 120  # seconds

# ruff: noqa: E402  # Avoid circular imports

from .agent import Agent, AgentConfig
from .agent_client import HTTPAgentClient, make_simple_agent_server
from .memory_agent import MemoryAgent
from .react_agent import ReActAgent
from .simple_agent import NoToolsSimpleAgent, SimpleAgent, SimpleAgentState
from .tree_of_thoughts_agent import TreeofThoughtsAgent

__all__ = [
    "DEFAULT_LLM_COMPLETION_TIMEOUT",
    "Agent",
    "AgentConfig",
    "HTTPAgentClient",
    "MemoryAgent",
    "NoToolsSimpleAgent",
    "ReActAgent",
    "SimpleAgent",
    "SimpleAgentState",
    "TreeofThoughtsAgent",
    "make_simple_agent_server",
]



from __future__ import annotations

import asyncio
import logging
import random
import typing
from collections import UserList
from collections.abc import Awaitable, Callable, Iterator
from enum import StrEnum, auto
from itertools import chain
from typing import Any, cast

import numpy as np
import torch
from pydantic import BaseModel, ConfigDict, Field
from tqdm import tqdm

from ldp.graph import eval_mode
from ldp.graph.async_torch import AsyncTorchModule

logger = logging.getLogger(__name__)


class ReplayBuffer(UserList[dict]):
    """A base replay buffer that only allows adding and sampling."""

    @staticmethod
    def _batched_iter(
        data: list[dict],
        batch_size: int,
        shuffle: bool = True,
        infinite: bool = False,
    ) -> Iterator[dict]:
        while True:
            indices = list(range(len(data)))
            if shuffle:
                random.shuffle(indices)

            for i in range(0, len(data), batch_size):
                keys = data[0].keys()

                batch: dict[str, list] = {k: [] for k in keys}
                for j in indices[i : i + batch_size]:
                    if data[j].keys() != keys:
                        raise RuntimeError(
                            "Found buffer element with inconsistent keys"
                        )

                    for k in keys:
                        batch[k].append(data[j][k])

                yield batch

            if not infinite:
                break

    def batched_iter(
        self,
        batch_size: int,
        shuffle: bool = True,
        infinite: bool = False,
    ) -> Iterator[dict]:
        return self._batched_iter(self.data, batch_size, shuffle, infinite)

    def resize(self, size: int) -> None:
        """Optional method for the buffer to resize itself."""

    async def prepare_for_sampling(self) -> None:
        """Optional method for the buffer to prepare itself before sampling."""

    @staticmethod
    def sample_from(*buffers: ReplayBuffer, **kwargs) -> Iterator[dict]:
        """Helper method to uniformly sample from multiple buffers."""
        if any(isinstance(b, PrioritizedReplayBuffer) for b in buffers):
            # This is because PrioritizedReplayBuffer determines samples inside
            # batched_iter, so we cannot rely on buffer.data.
            raise RuntimeError(
                "sample_from does not support prioritized replay buffers"
            )

        all_buffers = list(chain.from_iterable(b.data for b in buffers))
        return ReplayBuffer._batched_iter(data=all_buffers, **kwargs)


class CircularReplayBuffer(ReplayBuffer):
    def resize(self, size: int):
        if len(self) > size:
            self.data = self.data[-size:]


class RandomizedReplayBuffer(ReplayBuffer):
    def resize(self, size: int):
        if len(self) > size:
            self.data = random.sample(self.data, size)


class PrioritizedReplayBuffer(ReplayBuffer):
    """Implements a variant of https://arxiv.org/abs/1511.05952.

    Instead of updating the TD error on the fly, we compute it for all samples
    in the buffer before `update``. This allows us to efficiently
    batch prioritization and lets us sample w/o replacement.

    Also note that we define the TD error using the MC return, not the one-step
    return, for now. One-step may be possible using `next_state_action_cands`.

    As we expect our buffers to be O(100k) at most, we can afford to skip
    the binary heap implementation and just do a linear scan.
    """

    def __init__(
        self, alpha: float, ranked: bool, q_function: Callable[..., Awaitable]
    ):
        super().__init__()
        self.alpha = alpha
        self.ranked = ranked
        self.q_function = q_function

        self.buf_size: int | None = None

    def resize(self, size: int):
        self.buf_size = size

    @staticmethod
    async def _call_q(
        q_function: Callable[..., Awaitable], pbar: tqdm, *args, **kwargs
    ) -> float:
        # TODO: clean up this branching and force user to specify a Callable[..., Awaitable[float]]
        if isinstance(q_function, AsyncTorchModule):
            _, result = await q_function(*args, **kwargs)
        else:
            result = await q_function(*args, **kwargs)

        if isinstance(result, torch.Tensor):
            result = result.item()

        pbar.update()
        return cast("float", result)

    async def prepare_for_sampling(self):
        if self.buf_size is None:
            return

        pbar = tqdm(total=len(self.data), desc="Computing TD errors", ncols=0)
        async with eval_mode():
            values = await asyncio.gather(*[
                self._call_q(
                    self.q_function, pbar, *el["input_args"], **el["input_kwargs"]
                )
                for el in self.data
            ])
        for el, v in zip(self.data, values, strict=True):
            el["td_error"] = el["discounted_return"] - v

    def batched_iter(
        self,
        batch_size: int,
        shuffle: bool = True,
        infinite: bool = False,
    ) -> Iterator[dict]:
        if self.buf_size is None or (len(self.data) <= self.buf_size):
            # resize hasn't been called yet or we haven't hit the limit, so
            # use all samples
            buffer = self.data

        else:
            # roughly following Algo 1
            try:
                abs_tde = np.abs(
                    np.array([el["td_error"] for el in self.data])
                )  # L11-12
            except KeyError:
                raise RuntimeError(
                    "TD errors not available for all samples in the buffer. "
                    "Make sure to call prepare_for_update() after adding all samples "
                    "and before sampling."
                ) from None

            if self.ranked:
                ranks = np.argsort(abs_tde)
                prio = 1 / (ranks + 1)
            else:
                prio = abs_tde
            exp_prio = prio**self.alpha  # L9
            prob = exp_prio / exp_prio.sum()  # L9

            idcs = np.arange(len(self.data))
            sampled_idcs = np.random.choice(  # noqa: NPY002  # TODO: fix
                idcs, size=self.buf_size, p=prob, replace=False
            )
            buffer = [self.data[i] for i in sampled_idcs]

        return self._batched_iter(buffer, batch_size, shuffle, infinite)


class ReplayBufferType(StrEnum):
    # Maps to different buffer classes
    APPEND_ONLY = auto()
    CIRCULAR = auto()
    RANDOMIZED = auto()
    PRIORITIZED = auto()


class ReplayBufferConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    buf_type: ReplayBufferType = Field(
        default=ReplayBufferType.CIRCULAR, description="Circular is a good default."
    )
    size: int | None = None
    kwargs: dict[str, Any] = Field(default_factory=dict)

    def make_buffer(self, **kwargs) -> ReplayBuffer:
        # kwargs are only for prioritized case
        if self.buf_type == ReplayBufferType.APPEND_ONLY:
            return ReplayBuffer()
        if self.buf_type == ReplayBufferType.CIRCULAR:
            return CircularReplayBuffer()
        if self.buf_type == ReplayBufferType.RANDOMIZED:
            return RandomizedReplayBuffer()
        if self.buf_type == ReplayBufferType.PRIORITIZED:
            kwargs = self.kwargs | kwargs
            return PrioritizedReplayBuffer(**kwargs)
        typing.assert_never(self.buf_type)



"""This module defines various gradient estimators that can be patched in during backward passes."""

from __future__ import annotations

import logging
from functools import partial
from typing import Any

import tree

from .op_utils import CallID
from .ops import GradInType, OpCtx, OpResult, ResultOrValue

try:
    import torch

    from .torch_ops import TorchOp

except ImportError:
    torch = None  # type: ignore[assignment]

logger = logging.getLogger(__name__)


def assign_constant_grads(
    input_args: list[ResultOrValue],
    input_kwargs: dict[str, ResultOrValue],
    value: Any,
    descend: bool = True,
):
    if not descend:
        return [value] * len(input_args), dict.fromkeys(input_kwargs, value)

    # descend into nested objects
    arg_grads = [
        tree.map_structure(lambda _: value, OpResult.unwrap_value(arg))
        for arg in input_args
    ]
    kwarg_grads = {
        k: tree.map_structure(lambda _: value, OpResult.unwrap_value(v))
        for k, v in input_kwargs.items()
    }
    return arg_grads, kwarg_grads


def straight_through_estimator(
    ctx: OpCtx,  # noqa: ARG001
    input_args: list[ResultOrValue],
    input_kwargs: dict[str, ResultOrValue],
    grad_output: tree.Structure,
    call_id: CallID,  # noqa: ARG001
    descend: bool = True,
) -> GradInType:
    return assign_constant_grads(input_args, input_kwargs, grad_output, descend=descend)


def stop_grad(
    ctx: OpCtx,  # noqa: ARG001
    input_args: list[ResultOrValue],
    input_kwargs: dict[str, ResultOrValue],
    grad_output: tree.Structure,  # noqa: ARG001
    call_id: CallID,  # noqa: ARG001
) -> GradInType:
    # don't descend - want gradients to stop at the OpResult level
    return assign_constant_grads(input_args, input_kwargs, None, descend=False)


def zero_estimator(
    ctx: OpCtx,  # noqa: ARG001
    input_args: list[ResultOrValue],
    input_kwargs: dict[str, ResultOrValue],
    grad_output: tree.Structure,  # noqa: ARG001
    call_id: CallID,  # noqa: ARG001
) -> GradInType:
    """Sets the gradient of all inputs to zero.

    Note that this is not the same as truncating the compute graph (stop_grad),
    since upstream nodes can still optimize their logprobs. The zero estimator
    the unbiased choice if we have no information about the gradient.
    """
    return assign_constant_grads(input_args, input_kwargs, 0.0)


def llm_straight_through_estimator(
    ctx: OpCtx,  # noqa: ARG001
    input_args: list[ResultOrValue],
    input_kwargs: dict[str, ResultOrValue],
    grad_output: tree.Structure,
    call_id: CallID,  # noqa: ARG001
) -> GradInType:
    """Straight-through for an LLM: descend into the config, but not msgs/tools/tool_calls.

    See LLMCallOp.backward() for more details on this choice.
    Don't bother checking that input_args/input_kwargs have the right structure,
    since compute_grads() will raise if not.
    """
    config_grad = tree.map_structure(
        lambda _: grad_output, OpResult.unwrap_value(input_kwargs["config"])
    )
    grad_args = [grad_output] * len(input_args)
    grad_kwargs = {"config": config_grad}
    for arg in ("msgs", "tools", "tool_choice"):
        if arg in input_kwargs:
            grad_kwargs[arg] = grad_output

    return grad_args, grad_kwargs


def assign_default_grads(
    input_grads: GradInType,
    input_args: list[ResultOrValue],
    input_kwargs: dict[str, ResultOrValue],
    default_grad_val: float = 0.0,
) -> GradInType:
    """Sets a default value of default_grad_val for every element in input_grads.

    Example:
    - input_kwargs = {"a": {"b": 1, "c": 2}},
    - input_grad_kwargs = {"a": {"b": 0.1}}
    Output: input_grads[1] = {"a": {"b": 0.1, "c": default_grad_val}}

    Returns:
        GradInType: A tuple containing the updated input_grad_args and
            input_grad_kwargs with default values assigned where necessary.
    """

    def get_nested_value(data: tree.Structure, path: list) -> Any:
        """Traverse given path over data and return the value at the end of the path."""
        try:
            current_value = data
            for key in path:
                current_value = current_value[key]
        except (KeyError, IndexError):
            return None  # If path not found, return None (than default_grad_val will be assigned)
        else:
            return current_value

    def assign_default_gradients(
        input_grads: tree.Structure, path: list, _value: Any
    ) -> Any:
        """Assign default_grad_val where grads are missing."""
        return get_nested_value(input_grads, path) or default_grad_val

    input_args_kwargs = (input_args, input_kwargs)
    input_grads = tree.map_structure_with_path(
        partial(assign_default_gradients, input_grads),
        input_args_kwargs,
    )

    tree.assert_same_structure(input_grads, input_args_kwargs)
    return input_grads


class TorchParamBackwardEstimator:
    """
    Gradient estimator for `TorchOp` internal parameters.

    This estimator computes gradients with respect to the internal parameters of a
    `torch.nn.Module` by calling the `backward` method of the estimator instead of the default
    `backward` method of `TorchOp`. Computed gradients are stored in the context of the operation
    under the key `"grad_params"`.

    Examples:
        >>> torch_module = torch.nn.Sequential(
        ...     torch.nn.Linear(4, 4),
        ...     torch.nn.Linear(4, 1),
        ... )
        >>> torch_op = TorchOp(torch_module)
        >>> estimator = TorchParamBackwardEstimator(torch_module)
        >>> result = await torch_op(torch.randn(4, requires_grad=True))
        >>> result.compute_grads(backward_fns={"TorchOp": estimator.backward})

    Note:
        This estimator is only compatible with `TorchOp` operations.
    """

    def __init__(self, module: torch.nn.Module):
        if torch is None:
            raise RuntimeError(
                f"PyTorch library not found. Unable to use {type(self).__name__} class."
                " To install PyTorch dependencies, please run `pip install ldp[nn]`."
            )
        self.params = dict(module.named_parameters())

    def backward(
        self,
        ctx: OpCtx,
        input_args: list[ResultOrValue],
        input_kwargs: dict[str, ResultOrValue],
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        tensor_args, tensor_kwargs = ctx.get(call_id, TorchOp.CTX_TENSOR_INPUT_KEY)
        n_pos_args = len(tensor_args)
        n_pos_kwargs = len(tensor_kwargs)
        output = ctx.get(call_id, "output").value

        if not isinstance(grad_output, torch.Tensor):
            grad_output = torch.tensor(
                grad_output,
                dtype=output.dtype,
            )
        grad_output = grad_output.to(output.device)

        while grad_output.ndim < output.ndim:
            # Assume we can broadcast, so expand dims
            # e.g. if output.shape = (2, 1, 1) and grad_output is a scalar
            # then we want to expand to (1, 1, 1) and then broadcast
            grad_output = grad_output.unsqueeze(-1)

        if output.shape != grad_output.shape:
            raise RuntimeError(
                f"Output shape {output.shape} does not match grad_output shape"
                f" {grad_output.shape}"
            )

        gradients = torch.autograd.grad(
            output,
            [*tensor_args, *tensor_kwargs.values(), *self.params.values()],
            grad_outputs=grad_output,
        )

        grad_args = [grad.detach().cpu().float() for grad in gradients[:n_pos_args]]
        grad_kwargs = {
            k: grad.detach().cpu().float()
            for k, grad in zip(
                tensor_kwargs.keys(), gradients[n_pos_args:n_pos_kwargs], strict=True
            )
        }
        grad_params = {
            name: grad.detach().cpu().float()
            for name, grad in zip(
                self.params.keys(), gradients[n_pos_kwargs:], strict=True
            )
        }

        ctx.update(call_id=call_id, key="grad_params", value=grad_params)

        return grad_args, grad_kwargs



import asyncio
import random
import time
from contextlib import nullcontext
from uuid import UUID

import numpy as np
import pytest
import torch
from torch import nn

from ldp.graph import ConfigOp, FxnOp, PromptOp, compute_graph, set_training_mode
from ldp.graph.async_torch import (
    AsyncTorchModule,
    _get_autocast_context,  # noqa: PLC2701
    async_protect_torch_call,
)
from ldp.graph.gradient_estimators import straight_through_estimator as ste
from ldp.graph.torch_ops import TorchOp


@pytest.fixture(name="run_id")
def fixture_run_id() -> UUID:
    return UUID("12345678-1234-5678-1234-567812345678")


class AddModule(nn.Module):
    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return x + y


class MulModule(nn.Module):
    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return x * y


class SinModule(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.sin(x)


TEST_CASES = [
    pytest.param(
        AddModule(),
        (
            torch.tensor(2.0, requires_grad=True),
            torch.tensor(3.0, requires_grad=True),
        ),
        5.0,
        {"x": 1.0, "y": 1.0},
        id="addition",
    ),
    pytest.param(
        MulModule(),
        (
            torch.tensor(2.0, requires_grad=True),
            torch.tensor(3.0, requires_grad=True),
        ),
        6.0,
        {"x": 3.0, "y": 2.0},
        id="multiplication",
    ),
    pytest.param(
        SinModule(),
        (torch.tensor(np.pi / 4, requires_grad=True),),
        np.sin(np.pi / 4),
        {"x": np.cos(np.pi / 4)},
        id="non_linear",
    ),
]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("module", "inputs", "expected_output", "expected_grads"), TEST_CASES
)
@pytest.mark.parametrize("training_mode", [True, False])
async def test_torch_op(
    module, inputs, expected_output, expected_grads, training_mode: bool
):
    set_training_mode(training_mode)
    op = TorchOp(module)
    async with compute_graph():
        result = await op(*inputs)
    assert result.value.requires_grad == training_mode

    if isinstance(expected_output, torch.Tensor):
        assert torch.allclose(result.value, expected_output)
    else:
        assert result.value.detach() == pytest.approx(expected_output)

    raises_context = (
        nullcontext()
        if training_mode
        else pytest.raises(KeyError, match="key='tensor_input' not found")
    )

    with raises_context:
        grad_output = torch.ones_like(result.value)
        arg_grads, kwarg_grads = TorchOp.backward(  # noqa: RUF059
            op.ctx,
            [],
            {},  # NOTE: compute_grads() would fill this, but TorchOp.backwards() should ignore it
            grad_output=grad_output,
            call_id=result.call_id,
        )

        assert set(kwarg_grads.keys()) == set(expected_grads.keys())

        for param, expected in expected_grads.items():
            computed = kwarg_grads[param]
            if isinstance(expected, torch.Tensor):
                assert torch.allclose(torch.tensor(computed), expected)
            else:
                assert computed == pytest.approx(expected)


@pytest.mark.asyncio
async def test_with_kwargs():
    class ScaledAddModule(nn.Module):
        def forward(self, x, y, scale=1.0):
            return (x + y) * scale

    op = TorchOp(ScaledAddModule())
    x = torch.tensor(2.0, requires_grad=True)
    y = torch.tensor(3.0, requires_grad=True)
    scale = torch.tensor(2.0, requires_grad=True)

    async with compute_graph():
        result = await op(x, y, scale=scale)
    assert isinstance(result.value.detach(), torch.Tensor)
    assert result.value.detach() == pytest.approx(10.0)

    grad_output = torch.tensor(1.0)
    arg_grads, kwarg_grads = TorchOp.backward(
        op.ctx,
        [],
        {"x": x, "y": y, "scale": scale},
        grad_output,
        result.call_id,
    )

    assert not arg_grads
    assert len(kwarg_grads) == 3
    assert kwarg_grads["x"] == pytest.approx(2.0)  # d(result)/dx = scale = 2.0
    assert kwarg_grads["y"] == pytest.approx(2.0)  # d(result)/dy = scale = 2.0
    assert kwarg_grads["scale"] == pytest.approx(
        5.0
    )  # d(result)/d(scale) = x + y = 5.0


@pytest.mark.asyncio
async def test_torch_op_composition() -> None:
    # Define our ops
    config_op = ConfigOp(config={"scale": 2.0})
    # Don't use FURB118 here because we assert on kwargs below
    fxn_op = FxnOp(lambda x: x["scale"])  # noqa: FURB118

    class ScaleSum(nn.Module):
        def forward(self, x, y):
            return y * torch.sum(x)

    torch_op = TorchOp(ScaleSum())
    prompt_op = PromptOp("The result is: {result}")

    # Forward pass
    async with compute_graph():
        config = await config_op()
        fxn_result = await fxn_op(config)
        x = [1.0, 2.0, 3.0]
        torch_result = await torch_op(x, fxn_result)
        prompt_result = await prompt_op(result=torch_result)

    # Check forward pass results
    assert isinstance(torch_result.value, torch.Tensor)
    assert torch_result.value.detach() == pytest.approx(
        12.0
    )  # y * sum(x) = 2.0 * 6.0 = 12.0
    assert fxn_result.value == pytest.approx(2.0)
    assert prompt_result.value == "The result is: 12.0"

    # Backward pass
    loss_grad = -1.0  # Arbitrary d(loss)/d(result)
    prompt_result.compute_grads(
        loss_grad,
        backward_fns={
            PromptOp: ste,
            FxnOp: ste,
        },
    )
    fxn_grad = fxn_op.get_input_grads(fxn_result.call_id)
    torch_grad_args, torch_grad_kwargs = torch_op.get_input_grads(torch_result.call_id)

    # Check backward pass results

    # d(loss)/d(scale) = d(loss)/d(result) * d(result)/d(scale) = -1.0 * 6.0 = -6.0
    assert fxn_grad == ([], {"x": {"scale": -6.0}})

    assert not torch_grad_args
    # d(result)/dx = y = 2.0
    assert torch.allclose(torch_grad_kwargs["x"], torch.tensor([-2.0, -2.0, -2.0]))  # type: ignore[arg-type]
    # d(result)/dy = sum(x) = 6.0
    assert torch.allclose(torch_grad_kwargs["y"], torch.tensor(-6.0))  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_torch_concurrency():
    assert torch.is_grad_enabled()

    def check_no_grad(grad_expected: bool):
        # This takes the place of a pytorch module - we want to
        # see if torch.is_grad_enabled() is set correctly
        assert torch.is_grad_enabled() == grad_expected

    async def change_grads_a_lot():
        for _ in range(10):
            no_grad = random.choice([True, False])
            await asyncio.sleep(0)
            await async_protect_torch_call(check_no_grad, no_grad=no_grad)(  # type: ignore[arg-type]
                not no_grad
            )
            assert torch.is_grad_enabled()  # Make sure grad state is reset properly

            # The below is an example of doing unsafe concurrent operations in torch.
            # if no_grad:
            #     with torch.no_grad():
            #         asyncio.sleep(0.0)
            #         assert not torch.is_grad_enabled()
            # else:
            #     asyncio.sleep(0.0)
            #     check_no_grad(not no_grad)

    await asyncio.gather(*[change_grads_a_lot() for _ in range(10)])


@pytest.mark.asyncio
async def test_torch_autocast():
    model = nn.Linear(4, 4)

    def call_model():
        result = model(torch.rand(4))
        assert result.dtype == torch.bfloat16

    await async_protect_torch_call(
        call_model,  # type: ignore[arg-type]
        autocast_dtype=torch.bfloat16,
        autocast_device_type=torch.device("cpu").type,
    )()


class TestAsyncTorchModule:
    @pytest.mark.asyncio
    async def test_batching(self):
        model = torch.nn.Linear(2, 2)
        batch_size = 4
        max_wait = 1.0  # long timeout so we can measure it
        async_module = AsyncTorchModule(model, batch_size, max_wait)

        # First test that we can do a single call
        start = time.time()
        result = await async_module(input=torch.rand(2))
        assert max_wait < time.time() - start < max_wait * 2
        assert result.shape == (2,)

        # Now check that we can do a batched call
        start = time.time()
        results = await asyncio.gather(*[
            async_module(input=torch.rand(2)) for _ in range(batch_size)
        ])
        # should not have waited for timeout since we hit batch_size
        assert time.time() - start < max_wait
        assert len(results) == batch_size
        assert all(r.shape == (2,) for r in results)

        # Finally check an ueven number of calls
        results = await asyncio.gather(*[
            async_module(input=torch.rand(2)) for _ in range(2 * batch_size + 1)
        ])
        assert len(results) == 2 * batch_size + 1
        assert all(r.shape == (2,) for r in results)

    @pytest.mark.parametrize("batch_size", [4, 10])
    @pytest.mark.parametrize("max_wait", [0.01, 0.1, 1.0])
    @pytest.mark.asyncio
    async def test_ordering(self, batch_size: int, max_wait: float):
        # Make sure that we are actually getting the right results back, and that results
        # aren't jumbled

        model = torch.nn.Linear(2, 1, bias=False)
        model.weight.data.fill_(1.0)
        async_module = AsyncTorchModule(model, batch_size, max_wait)

        # Run 2 loops to make sure subsequent calls don't interfere
        for _ in range(2):
            inputs = [torch.rand(2) for _ in range(32)]
            outputs = await asyncio.gather(*[async_module(input=x) for x in inputs])

            for inp, out in zip(inputs, outputs, strict=True):
                assert torch.allclose(out, inp.sum())

            # Make sure we didn't leave any dangling work in the buffer
            assert not async_module._work_buffer

    # For tiny models, we expect a slowdown, but less than 3x.
    # As the model gets bigger, we get a slight speedup.
    # The real benefit is when the model fills up a GPU, but we don't want that in tests.
    @pytest.mark.skip(
        reason="The speedups are very hardware-dependent, so do not automatically run.",
    )
    @pytest.mark.parametrize(("dim", "expected_speedup"), [(2, 3), (4096, 0.8)])
    @pytest.mark.asyncio
    async def test_performance(self, dim: int, expected_speedup: float):
        batch_size = 10
        model = torch.nn.Linear(dim, dim)
        async_module = AsyncTorchModule(model, batch_size, 0.001)
        sync_batch = torch.rand(batch_size, dim)
        start = time.time()
        _ = model(sync_batch)
        sync_time = time.time() - start

        async_batch = list(sync_batch)
        start = time.time()
        _ = await asyncio.gather(*[async_module(input=x) for x in async_batch])
        async_time = time.time() - start
        print(
            f"Dimension: {dim}; torch.nn.Module: {sync_time:.6f}s; AsyncTorchModule:"
            f" {async_time:.6f}s"
        )

        assert async_time < sync_time * expected_speedup


@pytest.mark.parametrize("device_type", ["cpu", "cuda", "mps"])
def test_autocast(device_type: str):
    with _get_autocast_context(torch.bfloat16, device_type) as ctx:
        if device_type in {"mps", "cpu"}:
            # MPS and CPU devices don't support autocast
            assert not torch.is_autocast_enabled()
        else:
            # CUDA will only work if we've got a GPU
            assert torch.is_autocast_enabled() == torch.cuda.is_available()



import argparse
import asyncio
import pickle
from contextlib import suppress
from os import PathLike
from pathlib import Path

from aviary.core import Environment

from ldp.agent import Agent
from ldp.alg.callbacks import TerminalPrintingCallback
from ldp.alg.rollout import RolloutManager


def get_or_make_agent(agent: Agent | str | PathLike) -> Agent:
    if isinstance(agent, Agent):
        return agent

    if isinstance(agent, str):
        with suppress(KeyError):
            return Agent.from_name(agent)

    path = Path(agent)
    if not path.exists():
        raise ValueError(f"Could not resolve agent: {agent}")

    with path.open("rb") as f:
        return pickle.load(f)  # noqa: S301


def get_or_make_environment(environment: Environment | str, task: str) -> Environment:
    if isinstance(environment, Environment):
        return environment

    if isinstance(environment, str):
        with suppress(KeyError):
            return Environment.from_name(environment, task=task)

    raise ValueError(
        f"Could not resolve environment: {environment}. Available environments:"
        f" {Environment.available()}"
    )


async def main(
    task: str,
    environment: Environment | str,
    agent: Agent | str | PathLike = "SimpleAgent",
):
    agent = get_or_make_agent(agent)

    callback = TerminalPrintingCallback()
    rollout_manager = RolloutManager(agent=agent, callbacks=[callback])

    _ = await rollout_manager.sample_trajectories(
        environment_factory=lambda: get_or_make_environment(environment, task)
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("task", help="Task to prompt environment with.")
    parser.add_argument(
        "--env", required=True, help="Environment to sample trajectories from."
    )
    parser.add_argument(
        "--agent", default="SimpleAgent", help="Agent to sample trajectories with."
    )
    args = parser.parse_args()

    asyncio.run(main(args.task, args.env, args.agent))



import pytest

from ldp.alg.optimizer.replay_buffers import (
    CircularReplayBuffer,
    PrioritizedReplayBuffer,
)


@pytest.mark.asyncio
async def test_circular_buffer() -> None:
    buf = CircularReplayBuffer()

    samples = [{"state": 1, "action": 2, "reward": 3, "t": t} for t in range(5)]
    buf += samples
    buf.resize(3)  # should eject t=0, 1
    assert {sample["t"] for sample in buf} == {2, 3, 4}

    await buf.prepare_for_sampling()

    # check we can iterate
    next(buf.batched_iter(batch_size=3))

    # add a bad sample
    buf.append({})
    with pytest.raises(
        RuntimeError, match="Found buffer element with inconsistent keys"
    ):
        next(buf.batched_iter(batch_size=4))

    buf.clear()
    assert not buf, "Failed to clear data"


async def _dummy_q_function(*args, **kwargs) -> float:  # noqa: ARG001, RUF029
    return 1.0


@pytest.mark.asyncio
async def test_prioritized_buffer():
    buf = PrioritizedReplayBuffer(alpha=1, ranked=False, q_function=_dummy_q_function)

    buf += [
        {
            "input_args": (),
            "input_kwargs": {},
            "discounted_return": -1.0 if t % 2 else 1.0,
            "t": t,
        }
        for t in range(6)
    ]
    buf.resize(3)

    with pytest.raises(RuntimeError, match="TD errors not available"):
        next(buf.batched_iter(batch_size=3))

    await buf.prepare_for_sampling()

    # check we can iterate
    batch = next(buf.batched_iter(batch_size=3))

    # The odd timesteps should have priority because they have higher error
    assert all(t % 2 for t in batch["t"])



from __future__ import annotations

import json
import logging
import os
from collections.abc import Callable, Hashable, Iterable
from contextlib import suppress
from typing import TYPE_CHECKING, Any, ClassVar, Self, cast
from uuid import UUID

import aiofiles
from aviary.core import Message, ToolRequestMessage, ToolResponseMessage, join
from pydantic import BaseModel, ConfigDict, Field, JsonValue, field_validator

from ldp.graph import OpResult
from ldp.graph.op_utils import _lazy_import_networkx
from ldp.utils import discounted_returns

if TYPE_CHECKING:
    from aviary.core import Environment


logger = logging.getLogger(__name__)


class Transition(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    # Sentinel value for missing observation, as opposed to empty observation
    # Only used for tests; a user should never use this.
    NO_OBSERVATION: ClassVar[list[Message]] = []

    timestep: int = Field(description="Zero-indexed MDP timestep t.")

    agent_state: Any = Field(
        description=(
            "Agent.get_asv's input. This is `s_t` in the POMDP. Note that `s_0` comes"
            " from `Agent.init_state()`"
        )
    )
    next_agent_state: Any = Field(
        description="Agent.get_asv's output. This is s_t+1 in the POMDP."
    )

    observation: list[ToolResponseMessage | Message] = Field(
        description="Agent.get_asv's input. This is o_t in the POMDP."
    )
    next_observation: list[ToolResponseMessage | Message] = Field(
        description="Environment.step output. This is o_t+1 in the POMDP."
    )

    action: OpResult[ToolRequestMessage] | None = Field(
        default=None, description="Agent.get_asv output. This is a_t in the POMDP."
    )

    reward: float = Field(
        default=0.0, description="Environment.step output. This is r_t in the POMDP."
    )

    truncated: bool = Field(
        default=False, description="timestep t's Environment.step output."
    )
    done: bool = Field(
        default=False, description="timestep t's Environment.step output."
    )
    value: float = Field(
        default=0.0,
        description=(
            "Value estimate output from timestep t's Agent.get_asv. This is v(s_t)"
            " [state value function] or q(s_t, a_t) [state-action value]."
        ),
    )
    # JsonValue so we can serialize
    metadata: dict[str, JsonValue] = Field(default_factory=dict)

    @field_validator("action", mode="before")
    @classmethod
    def construct_action(
        cls, action: OpResult[ToolRequestMessage] | dict | None
    ) -> OpResult[ToolRequestMessage] | None:
        if isinstance(action, dict):
            return OpResult.from_dict(ToolRequestMessage, action)
        return action

    @property
    def failed(self) -> bool:
        """Get if an exception was encountered during rollout, for convenience.

        If True, this transition should not be trained on.
        Failed transitions are for debugging purposes.
        """
        return bool(self.metadata.get("exception"))

    def model_dump_json(self, *, indent: int | None = None, **kwargs) -> str:
        # TODO: decide if we should override Transition.model_dump() instead.
        # The kwargs for model_dump are the same as super().model_dump_json,
        # with the exception of indent.
        dump = self.model_dump(**kwargs)
        if self.action is not None:
            dump["action"] = self.action.to_dict()
        return json.dumps(dump, indent=indent)


class Trajectory(BaseModel):
    model_config = ConfigDict(extra="forbid")

    traj_id: str | None = None
    steps: list[Transition] = Field(default_factory=list)
    metadata: dict[str, JsonValue] = Field(
        default_factory=dict,
        description=(
            "Optional JSON metadata on the trajectory, for example the ID of a task"
            " this trajectory was run on."
        ),
    )

    @property
    def failed(self) -> bool:
        return any(step.failed for step in self.steps)

    @property
    def done(self) -> bool:
        if not self.steps:
            return False
        return self.steps[-1].done

    async def to_jsonl(self, filename: str | os.PathLike) -> None:
        async with aiofiles.open(filename, "w") as f:
            await f.write(json.dumps(self.traj_id) + "\n")
            for s in self.steps:
                await f.write(s.model_dump_json() + "\n")

    @classmethod
    async def from_env(cls, env: Environment, **kwargs) -> Self:
        """Create a Trajectory while propagating environment ID."""
        traj_metadata = kwargs.pop("metadata", {})
        with suppress(NotImplementedError, ValueError):
            traj_metadata["env_id"] = await env.get_id()
        return cls(metadata=traj_metadata, **kwargs)

    @classmethod
    def from_jsonl(cls, filename: str | os.PathLike) -> Self:
        with open(filename, encoding="utf-8") as f:
            reader = iter(f)
            traj = cls(traj_id=json.loads(next(reader)))
            for json_line in reader:
                data = json.loads(json_line)
                # logprob may have been serialized, but cannot be passed to
                # OpResult, so remove it here.
                with suppress(KeyError):
                    data["action"].pop("logprob")
                traj.steps.append(Transition(**data))
        return traj

    def compute_discounted_returns(self, discount: float = 1.0) -> list[float]:
        """Compute the discounted returns for each step in the trajectory."""
        return discounted_returns(
            rewards=[step.reward for step in self.steps],
            terminated=[step.truncated for step in self.steps],
            discount=discount,
        )


class TransitionTree:
    def __init__(self, root_id: str | UUID):
        """A tree of transitions.

        If A->B is an edge in this tree, then A and B are consecutive
        transitions in an LDP. Any path from the root node to a terminal
        node constitutes a complete LDP.

        A node may be assigned an arbitrary weight, which will be treated
        as a relative probability of sampling that node. For example, if
        A(weight=1) and B(w=2) are both children of the same node,
        then we treat B as twice as likely as A.

        Args:
            root_id: A unique identifier for the root node of the tree.
                All IDs of transitions added to this tree must begin with
                the same root_id.
        """
        self.root_id = str(root_id)

        nx = _lazy_import_networkx()
        self.tree = nx.DiGraph()  # the actual tree
        self.rev_tree = nx.DiGraph()  # the same as self.tree, but with reversed edges

        self._add_node(self.root_id, transition=None, weight=1.0)

    def _add_node(
        self, step_id: str, transition: Transition | None, weight: float
    ) -> None:
        self.tree.add_node(step_id, transition=transition, weight=weight)
        self.rev_tree.add_node(step_id)

    def _add_edge(self, parent_step_id: str, child_step_id: str) -> None:
        self.tree.add_edge(parent_step_id, child_step_id)
        self.rev_tree.add_edge(child_step_id, parent_step_id)

    def get_transition(self, step_id: str) -> Transition:
        if step_id == self.root_id:
            raise ValueError("Root node has no transition.")

        return cast("Transition", self.tree.nodes[step_id]["transition"])

    def get_weight(self, step_id: str) -> float:
        return cast("float", self.tree.nodes[step_id]["weight"])

    def add_transition(
        self, step_id: str, step: Transition, weight: float = 1.0
    ) -> None:
        """Add a transition to the tree.

        Args:
            step_id: A unique identifier for this node in the tree.
                The expected form of the step ID is "{parent step ID}:{step index}".
            step: The transition to add.
            weight: Weight of the transition. Defaults to 1.0.
        """
        root_id, *step_ids = step_id.split(":")
        assert root_id == self.root_id, (
            f"Step ID {step_id} does not start with root ID {self.root_id}"
        )
        assert step_ids, "Step ID cannot be the same as the root ID."
        # TODO: maybe this should be warning?
        assert step_id not in self.tree, (
            f"Step ID {step_id} already exists in the tree."
        )

        self._add_node(step_id, transition=step, weight=weight)

        parent_id = ":".join([root_id, *step_ids[:-1]])
        if parent_id in self.tree:
            self._add_edge(parent_id, step_id)

    def get_trajectories(self) -> list[Trajectory]:
        """Return a list of trajectories.

        Since each path from the root node to a terminal node defines
        a unique trajectory, N(terminal node) trajectories will be returned.
        The trajectory ID will be set to the ID of the terminal step.

        Note that we include failed and truncated trajectories; it is up to the
        caller to decide what to do them.

        Returns:
            All trajectories in this tree.
        """
        trajs = []
        step: Transition | None

        for step_id, step in self.tree.nodes(data="transition"):
            if not step:
                # root node
                continue

            is_terminal = (
                # check terminal conditions in increasing order of expense
                step.done
                or step.truncated
                or step.failed
                or self.tree.out_degree(step_id) == 0
            )

            if not is_terminal:
                continue

            # set the ID to the terminal node, which uniquely identifies the path
            traj = Trajectory(traj_id=step_id)
            # Build the trajectory up from a terminal node
            current_step: Transition | None = step
            current_step_id = step_id

            # Walk backwards towards the root (current_step=None)
            while current_step:
                traj.steps.append(current_step)

                parent_step_id, *extra = list(self.rev_tree.successors(current_step_id))
                assert not extra, f"Expected a single parent, but got {len(extra) + 1}"

                current_step_id = parent_step_id
                current_step = self.tree.nodes[parent_step_id]["transition"]

            # would've added things in reverse order, so fix that here
            traj.steps.sort(key=lambda x: x.timestep)
            trajs.append(traj)

        return trajs

    def assign_mc_value_estimates(self, discount_factor: float = 1.0) -> None:
        """Assign Monte Carlo state-action value estimates to each transition (in-place).

        Args:
            discount_factor: The discount factor to use when computing cumulative
                future rewards.
        """
        for step_id in _lazy_import_networkx().topological_sort(self.rev_tree):
            step: Transition | None = self.tree.nodes[step_id]["transition"]
            if step is None:
                continue

            if children := list(self.tree.successors(step_id)):
                # V_{t+1}(s') = sum_{a'} p(a'|s') * Q_{t+1}(s', a')
                # Here we assume p(a'|s') is uniform over the sampled actions.
                # TODO: don't make that assumption where a logprob is available
                weights = [self.get_weight(child_id) for child_id in children]
                steps = [self.get_transition(child_id) for child_id in children]
                v_tp1 = sum(
                    w * step.value for w, step in zip(weights, steps, strict=True)
                ) / sum(weights)
            else:
                v_tp1 = 0.0

            # Q_t(s_t, a_t) = r_{t+1} + gamma * V_{t+1}(s_{t+1})
            # (we are assuming the environment is deterministic)
            step.value = step.reward + discount_factor * v_tp1

    def compute_advantages(self) -> None:
        """Replace Transition.value with an advantage (in-place).

        A(s, a) = Q(s, a) - V(s), where V(s) is estimated as the
        average of Q(s, a') over all a' sampled at s.

        TODO: put this in Transition.metadata['advantage']. Not doing
        this right now due to implementation details in an optimizer.
        """
        state_values: dict[str, float] = {}

        for step_id in cast(
            "Iterable[str]", _lazy_import_networkx().topological_sort(self.tree)
        ):
            # topological sort means we will update a parent node in-place before
            # descending to its children

            step: Transition | None = self.tree.nodes[step_id]["transition"]
            if step is None:
                state_values[step_id] = 0.0
                continue

            # First, update V_t so that we can compute A_{t+1} for children
            children = [
                self.tree.nodes[child_id] for child_id in self.tree.successors(step_id)
            ]
            if children:
                state_action_values = [child["transition"].value for child in children]
                weights = [child["weight"] for child in children]
                state_values[step_id] = sum(
                    w * v for w, v in zip(weights, state_action_values, strict=True)
                ) / sum(weights)

            # Now compute A_t and replace Q_t with it in-place
            # Note that we are guaranteed at least one parent, since the `step is None`
            # check above should have caught the root node.
            parent_id, *extra = list(self.rev_tree.successors(step_id))
            assert not extra, "self.tree is not a tree!"
            step.value -= state_values[parent_id]
            # TODO: switch to the following, instead of overwriting step.value.
            # See docstring for explanation.
            # step.metadata["advantage"] = step.value - state_values[parent_id]

    def remove_nonterminal_branches(self) -> TransitionTree:
        """Creates a new tree with only branches that end in terminal states (done=True).

        TODO: refactor this to not use trajectories. See the note in merge_identical_nodes
        for reasoning.
        """
        new_tree = TransitionTree(self.root_id)
        for trajectory in self.get_trajectories():
            if not trajectory.done:
                continue

            traj_id_parts = cast("str", trajectory.traj_id).split(":")

            for step in trajectory.steps:
                step_id = ":".join(traj_id_parts[: step.timestep + 2])
                if step_id not in new_tree.tree:
                    # Traversing the tree by traversing trajectories means we may
                    # visit early nodes multiple times. Only add if we haven't visited
                    # already.
                    new_tree.add_transition(
                        step_id=step_id,
                        step=step,
                        weight=self.get_weight(step_id),
                    )

        return new_tree

    def merge_identical_nodes(
        self,
        agent_state_hash_fn: Callable[[Any], Hashable],
        observation_hash_fn: Callable[
            [list[ToolResponseMessage | Message]], Hashable
        ] = join,
        next_observation_hash_fn: Callable[
            [list[ToolResponseMessage | Message]], Hashable
        ] = join,
    ) -> TransitionTree:
        """Merge nodes with identical (state, observation, action)s. Returns a new tree.

        NOTE: the step IDs of nodes will lose their lineage after merging nodes. For example,
        the parent of ROOT:0:0 may not be ROOT:0 if ROOT:0 got merged with ROOT:1. Algorithms
        that rely on step IDs (like remove_nonterminal_branches) will not work as expected.

        Args:
            agent_state_hash_fn: A function that returns a hashable representation
                of the agent state of a transition.
            observation_hash_fn: A function that returns a hashable representation
                of the observation messages of a transition.
            next_observation_hash_fn: A function that returns a hashable representation
                of the next observation messages of a transition.
        """
        new_tree = TransitionTree(self.root_id)

        # step hash -> step ID
        seen_hash_to_step_id: dict[int, str] = {}
        # old step ID -> new step ID
        node_remap: dict[str, str] = {self.root_id: self.root_id}

        for step_id in _lazy_import_networkx().topological_sort(self.tree):
            step: Transition | None = self.tree.nodes[step_id]["transition"]
            if step is None:
                continue

            state_hash = agent_state_hash_fn(step.agent_state)

            if step.action is not None:
                tool_request_msg = step.action.value
                # NOTE: we are ignoring tool call ID in the comparison of tool requests.
                # Thus, the tool call ID is excluded from the hash, so we can't just
                # simply call str(tool_request_msg)
                action_str = (tool_request_msg.content or "") + " ".join(
                    str(tc) for tc in tool_request_msg.tool_calls
                )
            else:
                action_str = ""

            step_hash = hash((
                state_hash,
                action_str,
                # (s, a, o): works for deterministic envs
                observation_hash_fn(step.observation),
                # (s, a, o, o'): works for both deterministic and stochastic envs
                next_observation_hash_fn(step.next_observation),
            ))
            step_weight = self.get_weight(step_id)

            if step_hash in seen_hash_to_step_id:  # Seen: merge
                merged_step_id = node_remap[step_id] = seen_hash_to_step_id[step_hash]
                # Not sure if this is the fastest way to do this
                new_tree.tree.nodes[merged_step_id]["weight"] += step_weight
            else:  # Unseen: don't merge
                node_remap[step_id] = seen_hash_to_step_id[step_hash] = step_id
                parent_id = node_remap[":".join(step_id.split(":")[:-1])]

                # manually add transitions, since the step_id substring relationship
                # will be broken
                new_tree._add_node(step_id, transition=step, weight=step_weight)
                new_tree._add_edge(parent_id, step_id)

        return new_tree



from collections.abc import Iterable, Mapping
from typing import Any

from aviary.core import Message, ToolRequestMessage

from ldp.graph import FxnOp, OpResult, PromptOp, compute_graph
from ldp.llms import prepend_sys_and_append_sys

from .llm_call import ParsedLLMCallModule


class ThoughtModule:
    @staticmethod
    def _downcast_to_message(message: Message | ToolRequestMessage) -> Message:
        if isinstance(message, ToolRequestMessage):
            # Downcast into a normal Message if the LLM tried to call tools
            return Message(role=message.role, content=message.content)
        return message

    def __init__(
        self, llm_model: dict[str, Any], first_sys_prompt: str, second_sys_prompt: str
    ):
        self.first_sys_prompt_op = PromptOp(first_sys_prompt)
        self.second_sys_prompt_op = PromptOp(second_sys_prompt)
        self.package_msg_op = FxnOp(prepend_sys_and_append_sys)
        self.llm_call = ParsedLLMCallModule[Message](
            llm_model, parser=self._downcast_to_message
        )

    @compute_graph()
    async def __call__(
        self,
        messages: Iterable[Message],
        first_prompt_kwargs: Mapping[str, Any],
        second_prompt_kwargs: Mapping[str, Any],
    ) -> OpResult[Message]:
        packaged_msgs = await self.package_msg_op(
            messages,
            initial_sys_content=await self.first_sys_prompt_op(**first_prompt_kwargs),
            final_sys_content=await self.second_sys_prompt_op(**second_prompt_kwargs),
        )
        return (await self.llm_call(packaged_msgs))[0]  # type: ignore[arg-type]



from __future__ import annotations

import asyncio
import math
import random
from collections.abc import Sequence
from typing import Any, cast

from aviary.core import Environment, TaskDataset
from pydantic import BaseModel, ConfigDict, Field

from ldp.agent import Agent
from ldp.alg.optimizer import Optimizer
from ldp.data_structures import Trajectory
from ldp.graph import OpResult, eval_mode, train_mode
from ldp.shims import tqdm

from .callbacks import Callback, ClearContextCallback
from .rollout import RolloutManager


async def _run_eval_loop(
    dataset: TaskDataset,
    rollout_manager: RolloutManager,
    batch_size: int,
    num_iterations: int | None,
    max_rollout_steps: int | None,
    callbacks: Sequence[Callback],
    shuffle: bool = False,
    num_rollouts_per_env: int = 1,
) -> None:
    await asyncio.gather(*[callback.before_eval_loop() for callback in callbacks])

    if num_iterations is None:
        try:
            num_iterations = math.ceil(len(dataset) / batch_size)
        except TypeError:
            raise ValueError(
                "If num_iterations is not provided, the "
                "dataset must be finite and implement __len__."
            ) from None
    if not num_iterations:
        return

    with tqdm(
        desc="Evaluation Iterations", ncols=0, total=num_iterations, leave=False
    ) as pbar:
        # We use pbar.n as a counter for the number of training steps
        while pbar.n < num_iterations:
            for batch in dataset.iter_batches(batch_size, shuffle=shuffle):
                all_trajectories: list[Trajectory] = []

                for _ in range(num_rollouts_per_env):
                    trajectories = await rollout_manager.sample_trajectories(
                        environments=batch, max_steps=max_rollout_steps
                    )
                    all_trajectories.extend(trajectories)

                await asyncio.gather(*[
                    callback.after_eval_step(all_trajectories) for callback in callbacks
                ])
                pbar.update()

                if pbar.n == num_iterations:
                    break

    await asyncio.gather(*[callback.after_eval_loop() for callback in callbacks])


class EvaluatorConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    batch_size: int = 1
    num_eval_iterations: int | None = Field(
        default=None,
        description=(
            "Number of eval iterations. "
            "If not provided, will exhaust the dataset. "
            "If 0, will not run the eval loop. "
        ),
    )
    num_rollouts_per_env: int = Field(
        default=1,
        description=(
            "Number of rollouts to execute for each environment in the dataset."
        ),
    )
    max_rollout_steps: int | None = None
    catch_agent_failures: bool = True
    catch_env_failures: bool = True
    clear_ctx_at_each_iter: bool = False
    shuffle: bool = Field(default=False, description="Shuffles the evaluation dataset.")

    def make_rollout_manager(
        self, agent: Agent, callbacks: Sequence[Callback]
    ) -> RolloutManager:
        return RolloutManager(
            agent=agent,
            callbacks=callbacks,
            catch_agent_failures=self.catch_agent_failures,
            catch_env_failures=self.catch_env_failures,
        )


class Evaluator:
    def __init__(
        self,
        config: EvaluatorConfig,
        agent: Agent,
        dataset: TaskDataset,
        callbacks: Sequence[Callback] | None = None,
    ):
        self.config = config
        self.agent = agent
        self.dataset = dataset
        self.callbacks = callbacks or []
        if self.config.clear_ctx_at_each_iter:
            clear_cb = ClearContextCallback()
            self.callbacks = [*self.callbacks, clear_cb] if callbacks else [clear_cb]
        self.rollout_manager = self.config.make_rollout_manager(agent, self.callbacks)

    @eval_mode()
    async def evaluate(self, **kwargs) -> None:
        """Run the agent over the provided dataset in eval mode."""
        return await self.run(**kwargs)

    async def run(self, **kwargs) -> None:
        """Run the agent over the provided dataset.

        **kwargs can be used to override config settings.

        This method does not set training mode, so it can be used to collect
        trajectories for offline training.
        """
        eval_kwargs: dict[str, Any] = {
            "batch_size": self.config.batch_size,
            "num_iterations": self.config.num_eval_iterations,
            "max_rollout_steps": self.config.max_rollout_steps,
            "shuffle": self.config.shuffle,
            "num_rollouts_per_env": self.config.num_rollouts_per_env,
        }
        await _run_eval_loop(
            dataset=self.dataset,
            rollout_manager=self.rollout_manager,
            callbacks=self.callbacks,
            **(eval_kwargs | kwargs),
        )


class OnlineTrainerConfig(EvaluatorConfig):
    batch_size: int
    num_train_iterations: int = Field(
        ge=0,
        description=(
            "Number of iterations (at one batch per iteration) to process during"
            " training, and setting to 0 skips training."
        ),
    )
    update_every: int = Field(
        default=1,
        description="Number of training iterations between optimizer update calls.",
        ge=1,
    )
    eval_every: int | None = Field(
        None,
        description=(
            "If set, will repeatedly evaluate on the validation set after this many"
            " iterations. If unset (default), no evaluation is performed."
        ),
    )
    eval_before: bool = Field(
        default=True,
        description=(
            "If True (default), evaluate on the validation set before kicking off"
            " training."
        ),
    )
    clear_ctx_at_each_iter: bool = False


class OnlineTrainer:
    def __init__(
        self,
        config: OnlineTrainerConfig,
        agent: Agent,
        optimizer: Optimizer,
        train_dataset: TaskDataset,
        eval_dataset: TaskDataset | None = None,
        callbacks: Sequence[Callback] | None = None,
    ):
        if config.eval_every is not None and eval_dataset is None:
            raise ValueError("Must specify eval_dataset if eval_every is set")

        self.config = config
        self.agent = agent
        self.train_dataset = train_dataset
        self.eval_dataset = eval_dataset
        self.optimizer = optimizer
        self.callbacks = callbacks or []
        if self.config.clear_ctx_at_each_iter:
            clear_cb = ClearContextCallback()
            self.callbacks = [*self.callbacks, clear_cb] if callbacks else [clear_cb]
        self.rollout_manager = self.config.make_rollout_manager(
            agent=agent, callbacks=self.callbacks
        )

    async def train(self) -> None:
        if self.config.eval_before:
            await self.evaluate()

        with tqdm(
            desc="Training Iterations", ncols=0, total=self.config.num_train_iterations
        ) as pbar:
            # We use pbar.n as a counter for the number of training steps
            while pbar.n < self.config.num_train_iterations:
                for batch in self.train_dataset.iter_batches(
                    self.config.batch_size, shuffle=True
                ):
                    await self._training_step(pbar.n, batch)
                    pbar.update()  # Increment pbar.n by 1

                    if (
                        self.config.eval_every is not None
                        and pbar.n % self.config.eval_every == 0
                    ):
                        await self.evaluate()

                    if pbar.n == self.config.num_train_iterations:
                        break  # Will also break out of the outer while loop

        await self.evaluate()

    @eval_mode()
    async def evaluate(self, **kwargs) -> None:
        await _run_eval_loop(
            dataset=cast("TaskDataset", self.eval_dataset),
            rollout_manager=self.rollout_manager,
            batch_size=self.config.batch_size,
            num_iterations=self.config.num_eval_iterations,
            max_rollout_steps=self.config.max_rollout_steps,
            callbacks=self.callbacks,
            **kwargs,
        )

    @train_mode()
    async def _training_step(self, i_iter: int, envs: Sequence[Environment]) -> None:
        training_batch: list[Trajectory] = []
        all_trajectories: list[Trajectory] = []

        for _ in range(self.config.num_rollouts_per_env):
            trajectories = await self.rollout_manager.sample_trajectories(
                environments=envs, max_steps=self.config.max_rollout_steps
            )

            training_batch.extend(traj for traj in trajectories if not traj.failed)
            all_trajectories.extend(trajectories)

        await self._optimizer_step(i_iter, training_batch)

        await asyncio.gather(*[
            callback.after_train_step(all_trajectories) for callback in self.callbacks
        ])

    async def _optimizer_step(
        self, i_iter: int, training_batch: Sequence[Trajectory]
    ) -> None:
        for traj in training_batch:
            for step in traj.steps:
                # TODO: make this async
                # step.action is not None because we checked traj.failed above
                cast("OpResult", step.action).compute_grads()

        self.optimizer.aggregate(training_batch)

        if (i_iter + 1) % self.config.update_every == 0:
            await self.optimizer.update()

            await asyncio.gather(*[
                callback.after_update() for callback in self.callbacks
            ])


class OfflineTrainerConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    batch_size: int = Field(ge=1)
    update_every: int = Field(
        default=1,
        description="Number of training iterations to run before updating the model.",
        ge=0,
    )
    clear_ctx_at_each_iter: bool = False
    num_epochs: int = Field(default=1, ge=0)


class OfflineTrainer:
    def __init__(
        self,
        config: OfflineTrainerConfig,
        agent: Agent,
        optimizer: Optimizer,
        train_trajectories: list[Trajectory],
        callbacks: Sequence[Callback] | None = None,
    ):
        self.config = config
        self.agent = agent
        self.optimizer = optimizer
        # copy so we can shuffle
        self.train_trajectories = train_trajectories.copy()
        self.callbacks = callbacks or []
        if self.config.clear_ctx_at_each_iter:
            clear_cb = ClearContextCallback()
            self.callbacks = [*self.callbacks, clear_cb] if callbacks else [clear_cb]

    async def train(self) -> None:
        random.shuffle(self.train_trajectories)

        full_batch = len(self.train_trajectories) <= self.config.batch_size
        num_total_steps = self.config.num_epochs * (
            1
            if full_batch
            else (len(self.train_trajectories) // self.config.batch_size)
        )

        with tqdm(total=num_total_steps, desc="Training Iterations", ncols=0) as pbar:
            for _ in range(self.config.num_epochs):
                if full_batch:
                    # Separating out the full batch case lets the user run a single update()
                    # step even if train_trajectories is empty. This can be useful if the
                    # optimizer is pre-populated with offline training data, for example.
                    batch = self.train_trajectories

                    self.optimizer.aggregate(batch, show_pbar=True)
                    await self.optimizer.update()

                    await asyncio.gather(*[
                        callback.after_update() for callback in self.callbacks
                    ])

                    await asyncio.gather(*[
                        callback.after_train_step(batch) for callback in self.callbacks
                    ])
                    pbar.update()

                else:
                    for training_step, i_batch_start in enumerate(
                        range(0, len(self.train_trajectories), self.config.batch_size)
                    ):
                        batch = self.train_trajectories[
                            i_batch_start : i_batch_start + self.config.batch_size
                        ]

                        self.optimizer.aggregate(batch)

                        if (training_step + 1) % self.config.update_every == 0:
                            await self.optimizer.update()
                            await asyncio.gather(*[
                                callback.after_update() for callback in self.callbacks
                            ])

                        await asyncio.gather(*[
                            callback.after_train_step(batch)
                            for callback in self.callbacks
                        ])
                        pbar.update()



import asyncio
import operator
import random
from typing import TypeVar, cast
from uuid import UUID

import litellm
import numpy as np
import pytest
import tenacity
import tree
from aviary.core import DummyEnv, Message, Tool, ToolRequestMessage
from lmi import CommonLLMNames, LLMResult
from lmi import LiteLLMModel as LLMModel

from ldp.graph import (
    CallID,
    ConfigOp,
    EmbeddingOp,
    FxnOp,
    LLMCallOp,
    Op,
    OpCtx,
    OpResult,
    PromptOp,
    compute_graph,
    eval_mode,
    get_call_id,
    get_run_id,
    set_training_mode,
)
from ldp.graph.gradient_estimators import straight_through_estimator as ste
from ldp.graph.ops import GradInType, ResultOrValue, TOutput_co
from ldp.llms.prompts import append_to_sys


class StatefulFxnOp(FxnOp[TOutput_co]):
    async def forward(self, *args, **kwargs) -> TOutput_co:
        result = await super().forward(*args, **kwargs)
        self.ctx.update(get_call_id(), "observed", value=True)
        return result


@pytest.mark.asyncio
async def test_call_ids() -> None:
    async def fxn(x) -> int:
        await asyncio.sleep(x)
        return int(x > 0)

    op = StatefulFxnOp[int](fxn)
    # In this test, we want to make sure that op calls can't interfere with each
    # each other's compute graphs. So we launch two tasks, where the second
    # one should finish before the first, allowing for the possibility of a
    # clobber.
    xs = [1, 0.5]

    async def call_op(x) -> tuple[OpResult, UUID]:
        async with compute_graph():
            result = await op(x)
            return result, get_run_id()

    results = await asyncio.gather(*[call_op(x) for x in xs])

    for result, run_id in results:
        assert run_id == result.call_id.run_id, "Inconsistent run IDs"
        assert op.ctx.get(key="observed", call_id=result.call_id) is True


@pytest.mark.asyncio
async def test_FxnOp() -> None:
    def fxn(x) -> int:
        return int(x > 0)

    op = FxnOp(fxn)
    async with compute_graph():
        op_result = await op(1)
    assert op_result.value == 1
    assert op_result.logprob is None

    with pytest.raises(ValueError, match="No gradients"):
        op.get_input_grads(op_result.call_id)
    op_result.compute_grads(1.0, backward_fns={FxnOp: ste})
    assert op_result.grad == 1.0, "We didn't persist the starting gradient"
    assert op.get_input_grads(op_result.call_id) == ([], {"x": 1})


T = TypeVar("T")


@pytest.mark.parametrize(
    "op_return",
    [
        (1, int),
        (1.0, float),
        ("1", str),
        (Message(content="1"), Message),
        (ToolRequestMessage(content="1"), ToolRequestMessage),
    ],
)
@pytest.mark.parametrize("training", [True, False])
@pytest.mark.asyncio
async def test_opresult_typing(op_return: tuple[T, type[T]], training: bool) -> None:
    op_return_value, op_return_type = op_return
    op = FxnOp[op_return_type](lambda: op_return_value)  # type: ignore[valid-type]

    set_training_mode(training)
    async with compute_graph():
        op_result = await op()

    # Confirm that the OpResult's type matches the value's type
    assert isinstance(op_result, OpResult)
    assert isinstance(op_result.value, op_return_type)


class TestLLMCallOp:
    @pytest.mark.vcr
    @pytest.mark.asyncio
    async def test_cost_tracking(self) -> None:
        model_name = "gpt-3.5-turbo"

        class LLMCallingEnv(DummyEnv):
            """Showing how environments can use LiteLLM to track their own costs."""

            def __init__(self):
                self.total_cost = 0.0

            async def reset(self) -> tuple[list[Message], list[Tool]]:
                async def generate_story() -> str:
                    """Generate a story."""
                    response = await litellm.acompletion(
                        model=model_name,
                        messages=[
                            {"content": "Please write a 5 word story", "role": "user"}
                        ],
                    )
                    self.total_cost += litellm.completion_cost(response)
                    return response.choices[0].message.content

                self.state = type(self).State(
                    messages=[Message(content="Generate a story")]
                )
                self.tools = [Tool.from_function(generate_story)]
                return self.state.messages, self.tools

        env = LLMCallingEnv()
        obs, tools = await env.reset()
        config = {"name": model_name, "temperature": 0.1}
        llm_op = LLMCallOp()

        # Perform one step
        async with compute_graph():
            op_result = await llm_op(config, msgs=obs, tools=tools)
        await env.exec_tool_calls(cast("ToolRequestMessage", op_result.value))

        # LLMCallOp track cost using run context
        result = llm_op.ctx.get(op_result.call_id, "result")
        assert result.cost > 0

        # Environment tracks its internal costs
        assert env.total_cost > 0

    @pytest.mark.vcr
    @pytest.mark.asyncio
    async def test_empty_tools(self) -> None:
        llm_call_op = LLMCallOp()
        message_result = await llm_call_op(
            LLMModel.model_fields["config"].default_factory(),  # type: ignore[call-arg, misc]
            msgs=[Message(content="Hello")],
            tools=[],
        )
        assert isinstance(message_result.value, ToolRequestMessage)
        assert not message_result.value.tool_calls

    @pytest.mark.vcr
    @pytest.mark.asyncio
    @pytest.mark.parametrize("temperature", [0.0, 0.5, 1.0])
    async def test_compute_logprob(self, temperature) -> None:
        config = {
            "name": CommonLLMNames.OPENAI_TEST.value,
            "temperature": temperature,
            "logprobs": True,
        }
        llm_op = LLMCallOp()
        output = await llm_op(config, msgs=[Message(content="Hello")])
        logp = llm_op.ctx.get(output.call_id, "logprob")
        raw_logp = llm_op.ctx.get(output.call_id, "raw_logprob")
        if temperature == 0.0:
            assert logp == 1.0
        elif temperature == 1.0:
            assert logp == raw_logp
        else:
            assert logp is None

    @pytest.mark.vcr
    @pytest.mark.asyncio
    async def test_validation(self) -> None:
        config = {"name": CommonLLMNames.OPENAI_TEST.value, "num_retries": 1}
        validator = StatefulValidator()
        llm_op = LLMCallOp(response_validator=validator)
        await llm_op(config, msgs=[Message(content="Hello")])
        assert validator.counter == 2  # first attempt should have failed

        config = {"name": CommonLLMNames.OPENAI_TEST.value, "num_retries": 0}
        llm_op = LLMCallOp(response_validator=StatefulValidator())
        with pytest.raises(tenacity.RetryError, match="ResponseValidationError"):
            await llm_op(config, msgs=[Message(content="Hello")])


class StatefulValidator:
    def __init__(self):
        self.counter = 0

    def __call__(self, response: LLMResult) -> None:
        self.counter += 1
        if self.counter == 1:
            raise ValueError("Invalid response")


@pytest.mark.asyncio
async def test_simple_prompt_graph() -> None:
    config = ConfigOp(config={"name": "hello"})
    prompt = PromptOp("Hello, {name}! You are {age} years old.")

    async with compute_graph():
        c = await config()
        s = await prompt(c, age=20)
    assert s.value == "Hello, hello! You are 20 years old."

    my_loss_grad = -2.0
    s.compute_grads(my_loss_grad, backward_fns={PromptOp: ste})
    grad = prompt.get_input_grads(s.call_id)
    assert grad[1]["age"] == -2.0


@pytest.mark.vcr
@pytest.mark.asyncio
async def test_llm_call_graph() -> None:
    sys_prompt_op = PromptOp(
        "Respond by first planning your actions, then write code, "
        "inspect its effect, and reason about correctness"
    )
    user_prompt_op = PromptOp("What is the result of this math equation: {equation}?")

    package_msg_op = FxnOp(append_to_sys)
    config = {
        "name": "gpt-3.5-turbo-0125",
        "temperature": 0.1,
        "logprobs": True,
        "top_logprobs": 1,
    }
    config_op = ConfigOp(config=config)

    # Now forward pass
    my_equation = "2 + 2"
    async with compute_graph():
        sys_prompt = await sys_prompt_op()
        user_prompt = await user_prompt_op(equation=my_equation)
        package_msg = await package_msg_op(user_prompt, sys_prompt)
        c = await config_op()
        result = await (llm_op := LLMCallOp())(c, package_msg)
    assert result.value is not None
    assert len(result.value.content) > 10  # type: ignore[arg-type]

    output_grad = -2.0  # some grad accrued from result
    result.compute_grads([output_grad])

    # check some grads are present
    _, g = llm_op.get_input_grads(result.call_id)
    assert g["config"] == dict.fromkeys(config, 0.0)
    assert g["msgs"] == 0.0

    _, g = config_op.get_input_grads(c.call_id)
    assert not g  # config op has no inputs

    _, g = user_prompt_op.get_input_grads(user_prompt.call_id)
    assert g["equation"] == 0

    # get examples
    assert llm_op.get_examples()

    # now inference pass
    with eval_mode():
        async with compute_graph():
            sys_prompt = await sys_prompt_op()
            user_prompt = await user_prompt_op(equation=my_equation)
            package_msg = await package_msg_op(user_prompt, sys_prompt)
            c = await config_op()
            result = await (llm_op := LLMCallOp())(c, package_msg)
    assert result.value.content is not None
    assert len(result.value.content) > 10


@pytest.mark.asyncio
async def test_nested_op():
    """Test that we can have a forward function that calls another Op."""
    inner_op_a = FxnOp(lambda x: x + 1)
    inner_op_b = FxnOp(lambda x: x + 1)

    async def nested_op(x: ResultOrValue[int]) -> OpResult[int]:
        async with compute_graph():
            x = await inner_op_a(x)
            return await inner_op_b(x)

    result = await nested_op(1)
    assert result.value == 3

    input_args, input_kwargs = result.inputs
    assert not input_args
    assert input_kwargs

    result.compute_grads(1)
    assert result.grad == 1


class PickFirstOp(Op[int]):
    async def forward(self, *args: int) -> int:
        return args[0]

    @classmethod
    def backward(
        cls,
        ctx: OpCtx,
        input_args,
        input_kwargs,
        grad_output: tree.Structure,
        call_id: CallID,
    ) -> GradInType:
        return [grad_output] + [None] * (len(input_args) - 1), {}


@pytest.mark.asyncio
async def test_multiple_op_calls():
    op_a = FxnOp[int](lambda x: x)
    op_b = PickFirstOp()

    # compute graph: A -> B <- A
    async with compute_graph():
        n_samples = 2
        samples = await asyncio.gather(*[op_a(i) for i in range(n_samples)])
        selected = await op_b(*samples)
        assert len(samples) == n_samples

        run_id = get_run_id()
        op_a_call_ids = op_a.get_call_ids({run_id})
        assert len(op_a_call_ids) == n_samples
        assert selected.value == 0

    selected.compute_grads(10.0, backward_fns={FxnOp: ste})
    for call_id in op_a_call_ids:
        call_idx = op_a.ctx.get(call_id, "output").value
        if call_idx == 0:
            # first call - grads should've backproped
            assert op_a.get_input_grads(call_id)[1]["x"] == 10.0
        else:
            # second call - compute graph should have been pruned by
            # PickFirstOp
            with pytest.raises(
                ValueError, match=r"No gradients have been computed for .*"
            ):
                op_a.get_input_grads(call_id)


@pytest.mark.asyncio
async def test_branching_compute_graph():
    # The goal of this test is to make sure that gradients are properly aggregated
    # when the output of an op is consumed by multiple downstream ops. We expect
    # the gradients from each downstream op to be summed together.

    op_a = FxnOp[int](lambda x: x)
    op_b: FxnOp[int] = FxnOp(lambda x: x + random.randint(1, 10))
    op_c = FxnOp[int](lambda *x: sum(x))  # noqa: FURB111

    # compute graph: a -> b1, b2 -> c
    async with compute_graph():
        a = await op_a(3)
        b1 = await op_b(a)
        b2 = await op_b(a)
        c = await op_c(b1, b2)

    loss_grad = -5.0
    c.compute_grads(loss_grad, backward_fns={FxnOp: ste})

    # b1, b2, c receive only one copy of gradient
    for result in (c, b2, b2):
        assert result.grad == loss_grad

    # since a is used by two op calls, it should receive 2x the gradient:
    assert a.grad == loss_grad * 2

    # and 2x gradient should be passed back to the input
    assert op_a.get_input_grads(a.call_id)[1]["x"] == loss_grad * 2


@pytest.mark.asyncio
async def test_clear_contexts():
    """Test that we can clear the contexts of Ops."""
    op1 = FxnOp(lambda x: x)
    op1.set_name("op1")
    op2 = FxnOp(lambda x: x)
    op2.set_name("op2")
    async with compute_graph():
        # Test global clear
        await op2(await op1(1))
        assert len(op1.ctx.data) == len(op2.ctx.data) == 1
        OpCtx.clear_contexts()
        assert not op1.ctx.data
        assert not op2.ctx.data

        # Test global clear by op name
        await op2(await op1(1))
        OpCtx.clear_contexts(op_names=["op1"])
        assert not op1.ctx.data
        assert len(op2.ctx.data) == 1

        # Test instance clear
        op2.clear_ctx()
        assert not op2.ctx.data


@pytest.mark.asyncio
@pytest.mark.vcr
@pytest.mark.parametrize("dense_embedding_size", [0, 256, 512])
@pytest.mark.parametrize("sparse_embedding_size", [0, 32, 64])
@pytest.mark.parametrize("name", ["text-embedding-3-small", "text-embedding-3-large"])
async def test_embedding_op(name, dense_embedding_size, sparse_embedding_size) -> None:
    if dense_embedding_size == sparse_embedding_size == 0:
        return
    op = EmbeddingOp(
        dense_embedding=name,
        dense_embedding_dim=dense_embedding_size,
        sparse_embedding_dim=sparse_embedding_size,
    )
    async with compute_graph():
        op_result = await op("test text")
    assert isinstance(op_result.value, np.ndarray)
    assert op_result.value.shape[0] == dense_embedding_size + sparse_embedding_size
    op_result.compute_grads()
    assert op.get_input_grads(op_result.call_id) == ([], {"string_input": None})


@pytest.mark.asyncio
async def test_op_lookup():
    op_a = FxnOp[int](lambda: 1)
    op_a.set_name("op_a")

    # don't set name on these two to make sure default names
    # still map uniquely
    op_b = FxnOp[int](lambda x: 2 * x)
    op_c = FxnOp[int](operator.neg)

    async with compute_graph():
        a = await op_a()
        b = await op_b(a)
        c = await op_c(b)

    assert a.op == op_a
    assert b.op == op_b
    assert c.op == op_c
